import numpy as np
from .graph import Graph
from itertools import zip_longest


def process_data(data):
  '''Checks and processes the data for storage in Tensor

  Supported types for data - [int, float, list, np.ndarray]
  Elements in data should be float or be typecastable to float

  Args:
    data (int or float or list or np.ndarray): Data to be processed
  
  Returns:
    Processed data
  
  Raises:
    TypeError: If data or its elements aren't typecastable to float
    TypeError: If data is not instance of supported types
  '''
  supported_types = (int, float, list, np.ndarray)
  if type(data) in supported_types:
    if not isinstance(data, np.ndarray):
      data = np.array(data)
    try:
      data = data.astype(float)
    except ValueError:
      raise TypeError("Elements of data should be of type float or be typecastable to float")
  else:
    raise TypeError(f"Expected data of types {supported_types} instead got {type(data)}")
  return data

def unbroadcast_data(data, orig_data_shape, broadcasted_shape):
  ''' Unbroadcasts the data to its original shape

  If data(a np object) is broadcasted during an operation, then it is unbroadcasted here,
  where all axes where it was broadcasted are summed along those axes to
  give the original shape of the data. If broadcasted_shape is None, then the data is
  returned as is.

  Args:
    data (np.ndarray): Data to be unbroadcasted
    orig_data_shape (tuple): Original shape of data before broadcasting
    broadcasted_shape (tuple): Shape to which data has been broadcasted to
  
  Returns:
    Data that is unbroadcasted
  '''

  def get_axes_to_be_summed(orig_data_shape, broadcasted_shape):
    '''Returns the axes along which data has been broadcasted

    Given the original data shape and its broadcasted shape, it returns True along
    an axis if their dimensions don't match, else returns False if they match,
    meaning there has been no broadcasting along that axis.
    https://numpy.org/doc/stable/user/basics.broadcasting.html

    Args:
      orig_data_shape (tuple): Original shape of data before broadcasting
      broadcasted_shape (tuple): Shape to which data has been broadcasted to
    
    Returns:
      tuple of axes on which there's been broadcasting
    '''
    axes_to_be_summed = []
    zipped = list(zip_longest(tuple(reversed(broadcasted_shape)), tuple(reversed(orig_data_shape)), fillvalue=None))
    for dim, (dim_broadcasted, dim_orig) in enumerate(reversed(zipped)):
      if dim_broadcasted!=dim_orig:
        axes_to_be_summed.append(dim)
    return tuple(axes_to_be_summed)

  if broadcasted_shape is not None:
    axes_to_be_summed = get_axes_to_be_summed(orig_data_shape, broadcasted_shape)
    unbroadcasted_data = np.sum(data, axis=axes_to_be_summed)
  else:
    unbroadcasted_data = data
  return unbroadcasted_data

def get_graph():
  '''Returns graph that is in use and present in Graph.graph

  If Graph.graph is None, then the global graph _NG_GRAPH is used

  Returns:
    Graph object that is currently used
  '''
  if Graph.graph is None:
    from .. import _NG_GRAPH
    graph = _NG_GRAPH
  else:
    graph = Graph.graph
  return graph


class new_graph:
  '''Creates a Graph object

  Context Manager to create a new graph if required anywhere and under the
  circumstances where it shouldn't interfere with the global _NG_GRAPH

  After entering, Graph object created is set in Graph.graph. After exiting
  the Graph.graph is set back to None which implies that global _NG_GRAPH will
  be used
  '''
  def __enter__(self):
    Graph.graph = Graph()
  
  def __exit__(self, exc_type, exc_value, exc_traceback):
    Graph.graph = None


class no_track:
  '''Prevents tracking of Tensors

  Context Manager to prevent creation of a backward graph, when gradient calculation
  is not required, for ex when testing a model after training it, you don't need
  any backward pass

  On entering, graph.track is set to False to indicate no tracking and on exiting, it is
  set back to True

  Parameters:
    graph (Graph): The current graph in use
  '''
  def __init__(self):
    self.graph = get_graph()

  def __enter__(self):
    self.graph.track = False
  
  def __exit__(self, exc_type, exc_value, exc_traceback):
    self.graph.track = True


def _evaluate_grad_check(analytical_grads, calculated_grads, epsilon, print_vals):
  '''Evaluates the gradient check and indicates whether it has passed or not

  Calculates the distance between the analytical and calculated gradients and if it is
  less than epsilon, then it has passed else failed

  Args:
    analytical_grads (list of int or float): Gradients that are calculated analytically
      by wiggling the parameters
    calculated_grads (list of int or float): Gradients that are calulated through
      backpropagation
    epsilon (float): The amount by which params need to be wiggled
    print_vals (bool): True if distance and verdict needs to be printed
  
  Returns:
    Distance between analytical and calculated gradients
  '''
  dist = np.linalg.norm(analytical_grads-calculated_grads)/(np.linalg.norm(analytical_grads) + np.linalg.norm(calculated_grads))
  if print_vals:
    print("Gradient Check Distance:", dist)
    if dist<epsilon:
      print("Gradient Check PASSED")
    else:
      print("Gradient Check FAILED")
  return dist


def _wiggle_params(analytical_grads, calculated_grads, params, get_loss, epsilon):
  '''Changes the params value by epsilon and calculates the analytical gradient

  First to each element in params.data epsilon is added and loss is calculated, similarly
  2*epsilon is subtracted to get another loss and using these two analytical gradient is calculated
  and is appended to analytical_grads and the gradient in param is appended to calculated_grads

  Args:
    analytical_grads (list of int or float): Gradients that are calculated analytically
      by wiggling the parameters
    calculated_grads (list of int or float): Gradients that are calulated through
      backpropagation
    params (list of Tensor): All params that need to be wiggled
    get_loss: function that is used to calculate the loss
    epsilon (float): The amount by which params need to be wiggled
  '''
  for param in params:
    if param.requires_grad:
      if not(isinstance(param.grad, np.ndarray)):
        param.grad = np.array(param.grad)
      for idx in np.ndindex(param.shape):
        with no_track():
          param.data[idx]+=epsilon # PLUS
          loss1 = get_loss()
          param.data[idx]-=(2*epsilon) # MINUS
          loss2 = get_loss()
          param.data[idx]+=epsilon # ORIGINAL
        calculated_grads.append(param.grad[idx])
        analytical_grads.append((loss1.data-loss2.data)/(2*epsilon))
    param.zero_grad() # to prevent any side effects


def grad_check(model, inputs, targets, loss_fn, epsilon=1e-7, print_vals=True):
  '''Performs Gradient Check

  Implements Gradient Check, to make sure that backprop is calculating
  the right gradients. All the parameters in the model are checked.

  If distance between backprop gradients and numerical gradients is less
  than epsilon, then the gradients are proper, if not there is an issue
    
  Args:
    model (Model): The Neural Network to be evaluated
    inputs (Tensor): Input data(No need for complete data, only sample enough)
    targets (Tensor): Targets
    loss_fn (Loss): Loss Function
    epsilon (float): The amount by which params need to be wiggled Defaults to 1e-7
    print_vals (bool): True if distance and verdict needs to be printed
  
  Returns:
    Distance between analytical and calculated gradients
  '''
  params = model.parameters()
  analytical_grads = []
  calculated_grads = []

  for param in params:
    param.zero_grad()

  def get_loss():
    outputs = model(inputs)
    loss = loss_fn(outputs, targets)
    return loss

  with new_graph():
    loss = get_loss()
    loss.backward()
    _wiggle_params(analytical_grads, calculated_grads, params, get_loss, epsilon)

  analytical_grads = np.array(analytical_grads)
  calculated_grads = np.array(calculated_grads)
  return _evaluate_grad_check(analytical_grads, calculated_grads, epsilon, print_vals)


def fn_grad_check(fn, inputs, params, targets=None, loss_fn=None, epsilon=1e-7, print_vals=True, **kwargs):
  '''Performs Gradient Check for a function

  Implements Gradient Check for a function instead of a complete model
  Any params that are required to be gradient checked can be specified

  Args:
    fn: Function to be gradient checked
    inputs (list of Tensor): inputs to the function
    params (list of Tensor): the params whose data can be wiggled to get the gradients
    targets (Tensor): targets of the function
    loss_fn (Loss): loss_fn to evaluate the function
    epsilon (float): The amount by which params need to be wiggled Defaults to 1e-7
    print_vals (bool): True if distance and verdict needs to be printed
    **kwargs: Any kwargs to be passed to fn
  
  Returns:
    Distance between analytical and calculated gradients
  '''
  if loss_fn is None:
    from ..nn.loss import MSE
    loss_fn = MSE()
  analytical_grads = []
  calculated_grads = []

  for param in params:
    param.zero_grad()

  def get_loss(targets=targets):
    outputs = fn(*inputs, **kwargs)
    if targets is None:
      from .tensor import Tensor as tensor
      targets = tensor(np.ones(outputs.shape))
    loss = loss_fn(outputs, targets)
    return loss
  
  with new_graph():
    loss = get_loss()
    loss.backward()
    _wiggle_params(analytical_grads, calculated_grads, params, get_loss, epsilon)

  analytical_grads = np.array(analytical_grads)
  calculated_grads = np.array(calculated_grads)
  return _evaluate_grad_check(analytical_grads, calculated_grads, epsilon, print_vals)



from .utils import process_data, unbroadcast_data
from .ops import add, sub, mul, div, pow as _pow, transpose, sum as _sum, exp, dot, flatten, reshape


class Tensor:
  '''Wrapper around NumPy arrays
  
  Parameters:
    data (int or float or list or np.ndarray): The data to be stored or manipulated
    requires_grad (bool): Whether the Tensor requires gradient to be calculated or not
      Defaults to False
    requires_broadcasting (bool): Whether the Tensor needs to be broadcasted when some Operation
      is performed. Defaults to True. This attribute is present as there are some operations like
      Convolution for which the kernel shouldn't be broadcasted to inputs shape
    grad (np.ndarray): The gradient value of the Tensor. Defaults to 0 if requires_grad else None
    grad_fn: The function that is set in Operation.backward, that'll be executed during backward
      pass to set the gradient of the Tensor
  '''

  def __init__(self, data, requires_grad=False, requires_broadcasting=True):
    '''
    Args:
      data (int or float or list or np.ndarray): The data to be stored or manipulated
      requires_grad (bool): Whether the Tensor requires gradient to be calculated or not
        Defaults to False
      requires_broadcasting (bool): Whether the Tensor needs to be broadcasted when some Operation
        is performed. Defaults to True. This attribute is present as there are some operations like
        Convolution for which the kernel shouldn't be broadcasted to inputs shape
    
    Raises:
      TypeError: if data doesn't belong to (int or float or list or np.ndarray)
    '''
    self.data = data
    self.requires_grad = requires_grad
    self.requires_broadcasting = requires_broadcasting
    self.grad = 0. if requires_grad else None
    self.grad_fn = None
  
  def zero_grad(self):
    '''Resets the grad of the Tensor to the defaults
    '''
    self.grad = 0. if self.requires_grad else None
  
  def backward(self, upper_grad=1., retain_graph=False):
    '''Kicks off the backward pass to calculate gradients

    Starts the gradient calculation for the backward pass from the
    Tensor, by calling the backward method of its corresponding Node
      
    Args:
      upper_grad (int or float or list or np.ndarray): The gradient with which to start the
        gradient calculation. Shape of upper_grad and shape of Tensor must be the same.
        Defaults to 1 as usually backward is called on a loss Tensor that has a scalar value
      retain_graph (bool): If the graph should be retained after backward pass or should be reset.
        Auto-removal of Tensors from the graph, which happens when the gradients of all tensors of
        node's parents have been calculated will be turned off.
    
    Raises:
      ValueError: If called on a Tensor that doesn't have requires_grad
      ValueError: If shapes of upper_grad and Tensor doesn't match
    '''
    if not(self.requires_grad):
      raise ValueError("Only tensors who requires_grad can call backward")
    from .utils import get_graph
    graph = get_graph()
    upper_grad = process_data(upper_grad)
    if self.shape!=upper_grad.shape:
      raise ValueError("Shapes of grad and Tensor data must match!")
    self.accumulate_grad(upper_grad) # Setting the grad of the current Tensor by adding the upper_grad
    node = graph.get_node(self)
    node.backward(retain_graph)
    if not(retain_graph):
      graph.reset_graph() # tensors are auto-removed, this is just for redundancy / safety
  
  def _backward(self, node, retain_graph, calculate_grads=True):
    '''The essence of autograd, final gradient calculations for the Tensor is performed here

    The gradient of each child is taken as upper gradient, the backward_fn of the 
    Node of the Tensor is executed to set the grad_fn of Tensor.

    grad_fn is executed, the grad is then unbroadcasted, if Tensor has been broadcasted
    during the Operation. auto-removal of Tensor from the graph is performed when
    retain_graph is False

    Args:
      node (Node): The Node corresponding to the Tensor
      retain_graph (bool): Whether the graph needs to be retained or reset
      calculate_grads (bool): Whether gradients should be calculated or not,
        Defaults to True
    '''
    from .utils import get_graph
    graph = get_graph()
    for child in node.children:
      if self.requires_grad and calculate_grads:
        child.backward_fn(*[node.tens for node in child.parents])
        upper_grad = child.tens.grad
        grad = self.grad_fn(upper_grad)
        grad = unbroadcast_data(grad, self.shape, child.parent_broadcast_shape)
        grad = grad.reshape(self.shape)
        self.accumulate_grad(grad)
      if not(retain_graph) and child.are_parents_visited():
        graph.remove_tensor(child.tens)
    if not(retain_graph) and node.are_parents_visited():
      graph.remove_tensor(node.tens)
  
  def set_grad_fn(self, grad_fn):
    '''Sets the grad_fn for the Tensor

    If requires_grad is True, then Tensor.grad_fn is set to grad_fn else None

    Args:
      grad_fn: Function that is set during execution of Operation.backward
    '''
    self.grad_fn = grad_fn if self.requires_grad else None

  def __add__(self, other):
    '''Performs element wise addition of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be added with
    
    Returns:
      Tensor of the result
    '''
    return add(self, other)
  
  def __radd__(self, other):
    '''Performs element wise addition of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be added with
    
    Returns:
      Tensor of the result
    '''
    return add(other, self)
  
  def __sub__(self, other):
    '''Performs element wise subtraction of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be subtracted from
    
    Returns:
      Tensor of the result
    '''
    return sub(self, other)
  
  def __rsub__(self, other):
    '''Performs element wise subtraction of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be subtracted from
    
    Returns:
      Tensor of the result
    '''
    return sub(other, self)
  
  def __mul__(self, other):
    '''Performs element wise multiplication of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be multiplied with
    
    Returns:
      Tensor of the result
    '''
    return mul(self, other)
  
  def __rmul__(self, other):
    '''Performs element wise multiplication of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be multiplied with
    
    Returns:
      Tensor of the result
    '''
    return mul(other, self)
  
  def __truediv__(self, other):
    '''Performs element wise division of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be divided with
    
    Returns:
      Tensor of the result
    '''
    return div(self, other)
  
  def __rtruediv__(self, other):
    '''Performs element wise division of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be divided with
    
    Returns:
      Tensor of the result
    '''
    return div(other, self)
  
  def __pow__(self, other):
    '''Performs element wise power of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be raised to
    
    Returns:
      Tensor of the result
    '''
    return _pow(self, other)
  
  def __rpow__(self, other):
    '''Performs element wise power of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be raised to
    
    Returns:
      Tensor of the result
    '''
    return _pow(other, self)
  
  def __pos__(self):
    '''Performs unary plus on the Tensor
    
    Returns:
      Tensor of the result
    '''
    return (1*self)
  
  def __neg__(self):
    '''Performs unary minus on the Tensor
    
    Returns:
      Tensor of the result
    '''
    return (-1*self)
  
  def dot(self, other):
    '''Performs dot product of Tensor with another object

    Args:
      other (int or float or list or np.ndarray): The object that needs to be dotted with
    
    Returns:
      Tensor of the result
    '''
    return dot(self, other)
  
  def sum(self, axis=None):
    '''Performs sum of Tensor along an axis

    Args:
      axis (None or int): The axis along which it should be summed
    
    Returns:
      Tensor of the result
    '''
    return _sum(self, axis)
  
  def exp(self):
    '''Performs exponentiation on the Tensor

    Returns:
      Tensor of the result
    '''
    return exp(self)
  
  def flatten(self):
    '''Flattens the Tensor from any dimension to 1D

    Returns:
      Flattened Tensor
    '''
    return flatten(self)
  
  def reshape(self, new_shape):
    '''Reshapes the Tensor to the new shape

    Args:
      new_shape (tuple): The shape to which the Tensor should be reshaped to
    
    Returns:
      Reshaped Tensor
    '''
    return reshape(self, new_shape)
  
  def accumulate_grad(self, grad):
    '''Accumulates gradients for the Tensor

    Adds the gradient calculated with the overall gradient of the Tensor because
    if gradients are flowing into a Tensor from two different paths, they need to be
    summed up

    Args:
      grad (np.ndarray): The gradient to be added/accumulated
    '''
    self.grad+=grad
  
  @property
  def data(self):
    '''Returns the data present in the Tensor

    Returns:
      data (np.ndarray): Data in the Tensor
    '''
    return self._data
  
  @data.setter
  def data(self, data):
    '''Sets the data to the Tensor

    Args:
      data (int or float or list or np.ndarray): Data to be set
    
    Raises:
      TypeError: If data is not instance of (int or float or list or np.ndarray)
    '''
    self._data = process_data(data)

  @property
  def shape(self):
    '''Returns the shape of the Tensor

    Returns:
      Shape of data in the Tensor
    '''
    return self.data.shape
  
  @property
  def T(self):
    '''Performs the transpose of the Tensor

    Returns:
      Transpose of the Tensor
    '''
    return transpose(self)
  
  def __getitem__(self, *indices):
    '''Performs the slice of the data in the Tensor

    Args:
      index (tuple of int and/or slice): The indices to be sliced along or indexed
    
    Returns:
      Tensor with sliced data
    '''
    supported_types = (int, slice)
    for index in indices:
      if type(index) not in supported_types:
        raise TypeError(f"Expected index of {supported_types} instead got {type(index)}")
    return Tensor(self.data[indices], requires_grad=self.requires_grad)
  
  def __repr__(self):
    return f'Tensor({self.data}, requires_grad={self.requires_grad})'
  
  def __str__(self):
    return f'Tensor( {self.data},\n requires_grad={self.requires_grad},\n grad_fn={self.grad_fn},\n shape={self.shape} )\n'


class Node:
  '''Used as an abstraction to connect the tensors together and hold relationships

  Each Tensor is assigned a Node and this Node monitors all the incoming edges(parents)
  and the outgoing edges(children)

  Parameters:
    children (list of Node): List of all Nodes which uses the current Node
      as an operand in an Operation
    parents (list of Node): List of all Nodes(operands) that has resulted in the creation
      of current Node
    parent_broadcast_shape (tuple or None): If the parent needs to be broadcasted from one shape to
      another, then the final broadcasted shape of the parent is stored here.
      If they cannot be broadcasted, then it is None
    backward_fn (Operation.backward): Sets the grad_fn of Tensor(operand) involved in the Operation
    visited (bool) - If Node is visited or not
  '''

  def __init__(self, tens):
    '''
      Args:
        tens (Tensor) - The Tensor corresponding to the Node
    '''
    self.tens = tens
    self.children = []
    self.parents = []
    self.parent_broadcast_shape = None
    self.backward_fn = None
    self.visited = False
  
  def top_sort(self):
    '''Performs topological sort of all Nodes starting from current Node

    Sorts the graph topologically, to perform backward pass efficiently, so that all the children's
    is calculated before the current node's gradient is calculated.

    Sorting is done by first checking if all the children are visited, if they are, then the current
    node is added to sorted_tensors if not, then topological sort is performed on children
    '''
    sorted_tensors = []
    if self.are_children_visited():
      self.visited = True
      sorted_tensors.append(self.tens)
      for parent in self.parents:
        if not(parent.visited):
          sorted_tensors+=parent.top_sort()
    else:
      for child in self.children:
        if not(child.visited):
          sorted_tensors+=child.top_sort()
    return sorted_tensors
  
  def backward(self, retain_graph):
    '''Initiates backward pass starting from current Node

    This first visits all the children to make sure that they aren't included in
    sorted_tensors as they aren't required as backward pass is being initiated from the current
    node.

    Then it pops its corresponding Tensor from sorted_tensors (it is the first tensor) so that
    _backward can be called on it with calculate_grads=False, so that grads arent calculated for
    it, but allows flushing of all Tensors

    Next it topologically sorts all Tensors starting from current Node then the Node
    corresponding to the Tensor is retreived, which is marked as visited and the Tensor's
    backward pass is initiated.

    Args:
      retain_graph (bool): If the graph should be retained after backward pass or flushed
        after backward calculation
    '''
    from .utils import get_graph
    graph = get_graph()
    graph.reset_visited()
    self.visit_all_children() # this allows for gradient calculation from any intermediate node in the graph
    sorted_tensors = self.top_sort()
    graph.reset_visited()

    sorted_tensors.pop(0) # Remove the Tensor corresponding to the current node
    self.visited = True
    self.tens._backward(self, retain_graph, calculate_grads=False)

    for tens in sorted_tensors:
      node = graph.get_node(tens)
      node.visited = True
      tens._backward(node, retain_graph)

  def visit_all_children(self):
    '''Marks all children as visited
    '''
    for child in self.children:
      child.visited = True

  def are_children_visited(self):
    '''Checks if all children are visited

    Returns:
      True if all children are visited else False
    '''
    for child in self.children:
      if not(child.visited):
        return False
    return True
  
  def are_parents_visited(self):
    '''Checks if all parents are visited

    Returns:
      True if all parents are visited else False
    '''
    for parent in self.parents:
      if not(parent.visited):
        return False
    return True
  
  def add_child(self, other):
    '''Adds a child to the Node

    Args:
      other (Node): The child Node
    '''
    self.children.append(other)
  
  def add_parent(self, other):
    '''Adds a parent to the Node

    Args:
      other (Node): The parent Node
    '''
    self.parents.append(other)
  
  def __repr__(self):
    return f'Node({self.tens})'
  
  def __str__(self):
    return f'Node( \n{self.tens}\nbackward_fn: {self.backward_fn}\nvisited: {self.visited}\n )'


from .layers import Layer
from ..autograd.ops.operation import Operation
import numpy as np


# <------------RELU------------>
class ReLU(Layer, Operation):
  '''ReLU Layer
  '''
  def forward(self, inputs):
    '''Calculates ReLU of inputs

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of result
    '''
    inputs = self.get_tensors(inputs)
    return self.get_result_tensor(np.maximum(0, inputs.data), inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of the Tensor

    If element in data is greater than zero, its local gradient will be 1
    else 0

    Args:
      inputs (Tensor): Operand
    '''
    inputs.set_grad_fn(lambda ug:np.where(inputs.data>=0, 1, 0)*ug)

  def __repr__(self):
    return 'ReLU()'
  
  def __str__(self):
    return 'ReLU'


# <------------SIGMOID------------>
class Sigmoid(Layer, Operation):
  '''Sigmoid Layer
  '''
  def forward(self, inputs):
    '''Calculates Sigmoid of inputs

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of result
    '''
    inputs = self.get_tensors(inputs)
    return self.get_result_tensor(1/(1+np.exp(-inputs.data)), inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of the Tensor

    Args:
      inputs (Tensor): Operand
    '''
    result = 1/(1+np.exp(-inputs.data))
    inputs.set_grad_fn(lambda ug:(result*(1-result))*ug)

  def __repr__(self):
    return 'Sigmoid()'
  
  def __str__(self):
    return 'Sigmoid'


# <------------TANH------------>
class Tanh(Layer, Operation):
  '''Tanh Layer
  '''
  def forward(self, inputs):
    '''Calculates Tanh of inputs

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of result
    '''
    inputs = self.get_tensors(inputs)
    return self.get_result_tensor(np.tanh(inputs.data), inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of the Tensor

    Args:
      inputs (Tensor): Operand
    '''
    result = np.tanh(inputs.data)
    inputs.set_grad_fn(lambda ug:(1-np.power(result,2))*ug)
  
  def __repr__(self):
    return 'Tanh()'
  
  def __str__(self):
    return 'Tanh'


# <------------SOFTMAX------------>
class Softmax(Layer, Operation):
  '''Softmax Layer

  Parameters:
    axis (None or int or tuple of int): Axis along which it should be calculated
  '''
  def __init__(self, axis):
    '''
    Args:
      axis (None or int or tuple of int): Axis along which it should be calculated
    '''
    self.axis = axis

  def forward(self, inputs):
    '''Calculates Softmax of inputs

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of result
    '''
    inputs = self.get_tensors(inputs)
    result = self.calc_softmax(inputs.data, axis=self.axis)
    return self.get_result_tensor(result, inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of the Tensor

    Quite a tricky one, first the Jacobian of each of the slices along
    the specified axis of the result is taken, which is then dotted with the
    corresponding slice of the upper gradient

    Args:
      inputs (Tensor): Operand
    '''
    def softmax_grad(arr, ug_slices): # arr will always be 1d array
      local_grad = -np.broadcast_to(arr, (arr.size, arr.size))
      np.fill_diagonal(local_grad, 1+np.diagonal(local_grad))
      local_grad = local_grad*arr.reshape(arr.size, 1)
      result = np.dot(local_grad, ug_slices.pop(0))
      return result
    
    def get_ug_slices(arr, ug_slices):
      ug_slices.append(arr)

    def grad_backward(ug):
      result = np.apply_along_axis(self.calc_softmax, self.axis, inputs.data)
      ug_slices = []
      np.apply_along_axis(get_ug_slices, self.axis, ug, ug_slices)
      grads = np.apply_along_axis(softmax_grad, self.axis, result, ug_slices)
      return grads

    inputs.set_grad_fn(grad_backward)
  
  @staticmethod
  def calc_softmax(arr, axis=None):
    '''Calculates stable Softmax
    
    Args:
      arr (np.ndarray): Array whose Softmax is to be calculated
      axis (int or tuple of int): Axis along which to calculate the Softmax
        Defaults to None
    
    Returns:
      Softmax of the array
    '''
    exponentiated = np.exp(arr-np.max(arr, axis=axis, keepdims=True))
    sum_val = np.sum(exponentiated, axis=axis, keepdims=True)
    return exponentiated/sum_val
  
  def __repr__(self):
    return f'Softmax(axis={self.axis})'
  
  def __str__(self):
    return 'Softmax'


# <------------LEAKYRELU------------>
class LeakyReLU(Layer, Operation):
  '''LeakyReLU Layer
  '''
  def __init__(self, leak=0.01):
    '''
    Args:
      leak (float): leak value
    '''
    self.leak = leak

  def forward(self, inputs):
    '''Calculates LeakyReLU of inputs

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of result
    '''
    inputs = self.get_tensors(inputs)
    arr = inputs.data
    return self.get_result_tensor(np.where(arr>=0, arr, self.leak*arr), inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of the Tensor

    If element in data is greater than zero, its local gradient will be 1
    else will be leak value

    Args:
      inputs (Tensor): Operand
    '''
    inputs.set_grad_fn(lambda ug: np.where(inputs.data>=0, 1, self.leak)*ug)

  def __repr__(self):
    return f'LeakyReLU(leak={self.leak})'
  
  def __str__(self):
    return 'LeakyReLU'


import dill
from .model import Model


def save_model(fpath, model):
  '''Saves the model

  Saves the model by pickling it onto a file

  Args:
    fpath (str): Path in which to save the model
    model (Model): Model to be saved
  
  Raises:
    TypeError: if model isn't an instance of Model
  '''
  if not isinstance(model, Model):
    raise TypeError(f'Expected Model object, instead got {type(model)}')
  with open(fpath,'wb') as fp:
    dill.dump(model, fp)
    print(f'MODEL SAVED at {fpath}')


def load_model(fpath):
  '''Loads the model

  Args:
    fpath (str): Path from which to load the model
  
  Returns:
    Model object that is loaded
  '''
  with open(fpath,'rb') as fp:
    model = dill.load(fp)
    print(f'MODEL LOADED from {fpath}')
  return model


def get_batches(inputs, targets=None, batch_size=None):
  '''Returns batches of inputs and targets

  Split the inputs and their corresponding targets into batches for efficient
  training

  Args:
    inputs (Tensor): Inputs to be batched
    targets (Tensor): Targets to be batched. Defaults to None
    batch_size (int): Size of the batches. Defaults to None meaning batch_size
      will be same as number of examples
  
  Yields:
    Batches of inputs and their corresponding targets

  Raises:
    AssertionError: If first dimensions of inputs and targets don't match
    ValueError: If batch_size is greater than number of examples
    ValueError: If batch_size is negative
    ValueError: If batch_size is 0
  '''
  if targets is not None:
    assert inputs.shape[0]==targets.shape[0], '0th dim should be number of examples and should match'

  num_examples = inputs.shape[0]

  if batch_size is not None:
    if batch_size > num_examples:
      raise ValueError("Batch size cannot be greater than the number of examples")
    elif batch_size<0:
      raise ValueError("Batch size cannot be negative")
    elif batch_size==0:
      raise ValueError("Batch size cannot be zero")
  else:
    batch_size = num_examples

  start = 0
  while start<num_examples:
    end = start+batch_size if start+batch_size<num_examples else num_examples
    if targets is not None:
      yield inputs[start:end], targets[start:end]
    else:
      yield inputs[start:end]
    start = end


import numpy as np
from copy import deepcopy
from ..autograd import tensor, dot
from ..autograd.ops import conv2d, conv3d, maxpool2d, maxpool3d
from ..autograd.ops.operation import Operation


class Container:
  '''Contains many Layers

  Parameters:
    eval (bool): Whether the Container is in eval mode
    layers (list of Layer/Container): Layer to be included in the container
  '''
  def __init__(self):
    self.eval = False
    self.layers = None

  def __call__(self, inputs):
    '''Calls the forward method

    Args:
      inputs (Tensor): Inputs to the container
    
    Returns:
      Tensor of the result
    '''
    return self.forward(inputs)

  def parameters(self, as_dict=False):
    '''Recursively goes through all the layers in the Container and gets the params of each Layer

    Args:
      as_dict (bool): Whether params need to be returned as a dict,
        Defaults to False
    
    Returns:
      list of Params
    '''
    params = []
    for layer in self.layers:
      if as_dict:
        params.append(layer.parameters(as_dict))
      else:
        params+=layer.parameters(as_dict)
    return params
  
  def set_eval(self, eval):
    '''Sets eval

    Sets its eval to the eval argument also recursively sets the eval
    of its layers

    Args:
      eval (bool): Whether Container is in eval mode or not
    '''
    self.eval = eval
    for layer in self.layers:
      layer.set_eval(eval)
  
  def set_params(self, container_params):
    '''Sets the params for all the layers
    '''
    for layer_param, layer in zip(container_params, self.layers):
      layer.set_params(layer_param)
  
  def freeze(self):
    '''Freezes all the layers present in the Container
    '''
    for layer in self.layers:
      layer.freeze()
  
  def unfreeze(self):
    '''Unfreezes all the layers present in the Container
    '''
    for layer in self.layers:
      layer.unfreeze()
  
  def __repr__(self):
    layers = []
    for layer in self.layers:
      layers.append(f'{layer.__str__()}')
    layers_repr = ', '.join(layers)
    return layers_repr
  
  def __str__(self):
    layers = []
    for layer in self.layers:
      layers.append(f'{layer.__repr__()}')
    layers_str = ', '.join(layers)
    return layers_str


class Layer:
  '''
    Fundamental building block of the model
  
  Parameters:
    eval (bool): Whether the Container is in eval mode
  '''
  def __init__(self):
    self.eval = False

  def __call__(self, inputs):
    '''Calls the forward method

    Args:
      inputs (Tensor): Inputs to the container
    
    Returns:
      Tensor of the result
    '''
    return self.forward(inputs)

  def parameters(self, as_dict=False):
    '''Returns the parameters in the Layer

    If any of the attributes in a Layer is instance of Param, then it is automatically
    considered as a param for the model

    Args:
      as_dict (bool): Whether params need to be returned as a dict,
        Defaults to False
    
    Returns:
      list of Params or dict
    '''
    params = {}
    for attr, val in self.__dict__.items():
      if isinstance(val, Param):
        params[attr] = val.data if as_dict else val
    return params if as_dict else list(params.values())
  
  def set_eval(self, eval):
    '''Sets eval

    Sets its eval to the eval argument

    Args:
      eval (bool): Whether Container is in eval mode or not
    '''
    self.eval = eval
  
  def set_params(self, layer_params):
    '''Sets the params for the current layer
    '''
    for attr, param_data in layer_params.items():
      param = self.__getattribute__(attr)
      param.data = param_data
  
  def freeze(self):
    '''Freezes all the Params in the Layer
    '''
    for param in self.parameters(as_dict=False):
      param.freeze()
  
  def unfreeze(self):
    '''Unfreezes all the Params in the Layer
    '''
    for param in self.parameters(as_dict=False):
      param.unfreeze()
  
  def __getstate__(self):
    '''Returns the state for the object that is to be pickled

    The instances of Param data are set to 0, to prevent saving of all the params
    in the Layer.
    If required, then the weights can be saved and loaded separately

    Returns:
      state of the current Layer
    '''
    state = deepcopy(self.__dict__)
    for param_attr in self.parameters(as_dict=True).keys():
      state[param_attr].data = 0 # Wanted to set it to None, but it isnt supported by Tensor, so set it to the next best 0
    return state
  
  def __setattr__(self, attr, val):
    '''Sets attributes for the Layer

    Doesn't allow modification/re-assigning of Param attributes

    Args:
      attr (str): Attribute to set the value to
      val (object): The value to be set to

    Raises:
      AttributeError: If a Param attr has already been defined/assigned
    '''
    if (isinstance(val, Param)) and (attr in self.__dict__):
      raise AttributeError(f"Attribute {attr} has already been defined, it cannot be defined again for a Param")
    object.__setattr__(self, attr, val)


class Param(tensor):
  '''Alias for Tensor

  Just an alias for Tensor, so that when params are gathered for a Layer, only these
  are automatically considered for param, while ignoring some helper Tensors which aren't
  necessarily param

  Parameters:
    __frozen (bool): Whether current Param is frozen or not. This is required because we
      need to know if it has been frozen before unfreeze is called
  '''

  def __init__(self, data, requires_grad=False, requires_broadcasting=True):
    super().__init__(data, requires_grad, requires_broadcasting)
    self.__frozen = False
  
  def freeze(self):
    '''Sets requires_grad=False
    '''
    self.requires_grad = False
    self.__frozen = True
  
  def unfreeze(self):
    '''Sets requires_grad=True only if its frozen

    frozen condition is checked because only if it was previously frozen, we can set requires_grad
    = True, if we don't check for it and requires_grad is False originally, then we might set it to
    True, which would be incorrect
    '''
    if self.__frozen:
      self.requires_grad = True
    self.__frozen = False
  
  def __str__(self):
    return f'Param({super().__str__()})'
  
  def __repr__(self):
    return f'Param({super().__repr__()})'


class Sequential(Container):
  '''Sequential Container
  
  Outputs of one layer are passed as inputs to the next layer, sequentially
  '''
  
  def __init__(self, *args):
    self.layers = args
  
  def forward(self, inputs):
    '''Forward pass of Sequential

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of the result
    '''
    for layer in self.layers:
      output = layer(inputs)
      inputs = output
    return output
  
  def __str__(self):
    return f'Sequential(\n{super().__str__()}\n)'
  
  def __repr__(self):
    return f'Sequential(\n{super().__repr__()}\n)'


class Linear(Layer):
  '''Implements a fully connected Layer

  Parameters:
    num_in (int): Number of inputs to the Layer
    num_out (int): Number of outputs from the Layer
    weights (Param): Weights of the Layer
    bias (Param): Bias of the Layer
  '''
  def __init__(self, num_in, num_out):
    self.num_in = num_in
    self.num_out = num_out
    self.weights = Param(np.random.randn(num_in, num_out), requires_grad=True)
    self.bias = Param(np.zeros((1, num_out)), requires_grad=True)
  
  def forward(self, inputs):
    '''Forward pass of Linear

    The inputs are dotted with weights and then bias is added

    Args:
      inputs (Tensor): Inputs to the Linear
    
    Returns:
      Tensor of the result
    '''
    return dot(inputs, self.weights) + self.bias
  
  def __repr__(self):
    return f'Linear({self.num_in}, {self.num_out})'
  
  def __str__(self):
    return f'Linear in:{self.num_in} out:{self.num_out}'


class Dropout(Layer, Operation):
  '''Dropout Layer
  
  https://youtu.be/D8PJAL-MZv8
  
  Parameters:
    prob (float): Probability with which to keep the inputs
      With probability=prob, the units are kept and with probability=1-prob,
      they are shut off
  '''
  def __init__(self, prob):
    Layer.__init__(self)
    assert prob>0 and prob<=1, 'Probability should be between 0 and 1'
    self.prob = prob
  
  def forward(self, inputs):
    '''Forward pass of Dropout

    The inputs are turned on with the given prob
    If in eval mode, then all inputs are always on

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of the result
    '''
    if self.eval:
      filter = np.ones(inputs.shape) # dont discard anything, just dummy because if eval or not, backward needs to have filter arg
    else:
      filter = np.where(np.random.random(inputs.shape)<self.prob, 1, 0)
    inputs, filter = self.get_tensors(inputs, filter)
    if not(self.eval): # Deliberately repeated condition check for eval, because if in eval, it shouldnt be scaled by prob
      result = (inputs.data*filter.data)/self.prob
    else:
      result = inputs.data
    return self.get_result_tensor(inputs.data, inputs, filter)
  
  def backward(self, inputs, filter):
    '''Sets the grad_fn of inputs only because filter doesnt have requires_grad=True

    Just multiplies the upper gradient with the filter set during forward pass and scales
    it by the probability

    Args:
      inputs (Tensor): Inputs to the Layer
      filter (Tensor): The dropout filter that was applied during forward pass
    '''
    if not(self.eval):
      inputs.set_grad_fn(lambda ug:(ug*filter.data)/self.prob)
    inputs.set_grad_fn(lambda ug:ug)
  
  def __repr__(self):
    return f'Dropout(prob={self.prob})'
  
  def __str__(self):
    return f'Dropout(prob={self.prob})'


class Conv2D(Layer):
  '''Implements Conv2D

  Parameters:
    padding (int): Padding value to be applied. Defaults to 0
    stride (int): Stride to be taken. Defaults to 1
    weights (Param): Kernel for the Convolution
    bias (Param): Bias for the Convolution
  
  Raises:
    ValueError: If kernel_shape isn't 2D tuple
  '''
  def __init__(self, kernel_shape, padding=0, stride=1):
    '''
    Args:
      kernel_shape (tuple of int): Shape of the kernel
    '''
    self.padding = padding
    self.stride = stride
    if len(kernel_shape)!=2:
      raise ValueError("Kernel shape can only have 2 dims")
    self.weights = Param(np.random.randn(*kernel_shape), requires_grad=True, requires_broadcasting=False)
    self.bias = Param(0, requires_grad=True, requires_broadcasting=False)
  
  def forward(self, inputs):
    '''Forward pass of Conv2D

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of the result
    '''
    return conv2d(inputs, self.weights, self.bias, self.padding, self.stride)
  
  def __repr__(self):
    return f'Conv2D(kernel_shape={self.weights.shape}, padding={self.padding}, stride={self.stride})'
  
  def __str__(self):
    return f'Conv2D(kernel_shape={self.weights.shape}, padding={self.padding}, stride={self.stride})'


class Conv3D(Layer):
  '''Implements Conv3D

  Parameters:
    padding (int): Padding value to be applied. Defaults to 0
    stride (int): Stride to be taken. Defaults to 1
    weights (Param): Kernel for the Convolution
    bias (Param): Bias for the Convolution
  
  Raises:
    ValueError: If kernel_shape isn't 2D tuple
  '''
  def __init__(self, in_channels, out_channels, kernel_shape, padding=0, stride=1):
    '''
    Args:
      in_channels (int): Number of channels in the inputs
      out_channels (int): Number of channels in the outputs
      kernel_shape (tuple of int): Shape of the kernel
    '''
    self.padding = padding
    self.stride = stride
    if len(kernel_shape)!=2:
      raise ValueError("Kernel shape can only have 2 dims")
    self.weights = Param(np.random.randn(out_channels, in_channels, *kernel_shape), requires_grad=True, requires_broadcasting=False)
    self.bias = Param(np.zeros(out_channels), requires_grad=True, requires_broadcasting=False)
  
  def forward(self, inputs):
    '''Forward pass of Conv3D

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of the result
    '''
    return conv3d(inputs, self.weights, self.bias, self.padding, self.stride)
  
  def __repr__(self):
    kernel_shape = self.weights.shape
    return f'Conv3D(out_channels={kernel_shape[0]}, in_channels={kernel_shape[1]}, kernel_shape={kernel_shape[2:]}, padding={self.padding}, stride={self.stride})'
  
  def __str__(self):
    kernel_shape = self.weights.shape
    return f'Conv3D(out_channels={kernel_shape[0]}, in_channels={kernel_shape[1]}, kernel_shape={kernel_shape[2:]}, padding={self.padding}, stride={self.stride})'


class MaxPool2D(Layer):
  '''Implements MaxPool2D

  Parameters:
    padding (int): Padding value to be applied. Defaults to 0
    stride (int): Stride to be taken. Defaults to 1
    kernel_shape (tuple of int): Shape of the kernel
  
  Raises:
    ValueError: If kernel_shape isn't 2D tuple
  '''
  def __init__(self, kernel_shape, padding=0, stride=1):
    self.padding = padding
    self.stride = stride
    if len(kernel_shape)!=2:
      raise ValueError("Kernel shape can only have 2 dims")
    self.kernel_shape = kernel_shape
  
  def forward(self, inputs):
    '''Forward pass of MaxPool2D

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of the result
    '''
    return maxpool2d(inputs, self.kernel_shape, self.padding, self.stride)
  
  def __repr__(self):
    return f'MaxPool2D(kernel_shape={self.kernel_shape}, padding={self.padding}, stride={self.stride})'
  
  def __str__(self):
    return f'MaxPool2D(kernel_shape={self.kernel_shape}, padding={self.padding}, stride={self.stride})'


class MaxPool3D(Layer):
  '''Implements MaxPool2D

  Parameters:
    padding (int): Padding value to be applied. Defaults to 0
    stride (int): Stride to be taken. Defaults to 1
    kernel_shape (tuple of int): Shape of the kernel
  
  Raises:
    ValueError: If kernel_shape isn't 2D tuple
  '''
  def __init__(self, kernel_shape, padding=0, stride=1):
    self.padding = padding
    self.stride = stride
    if len(kernel_shape)!=2:
      raise ValueError("Kernel shape can only have 2 dims")
    self.kernel_shape = kernel_shape
  
  def forward(self, inputs):
    '''Forward pass of MaxPool3D

    Args:
      inputs (Tensor): Inputs to the Layer
    
    Returns:
      Tensor of the result
    '''
    return maxpool3d(inputs, self.kernel_shape, self.padding, self.stride)
  
  def __repr__(self):
    return f'MaxPool3D(kernel_shape={self.kernel_shape}, padding={self.padding}, stride={self.stride})'
  
  def __str__(self):
    return f'MaxPool3D(kernel_shape={self.kernel_shape}, padding={self.padding}, stride={self.stride})'


import hickle as hkl
from itertools import chain as list_flattener
from .layers import Container, Layer
from ..autograd.utils import get_graph


class Model:
  def __call__(self, inputs):
    '''Abstracts the forward method

    Args:
      inputs (Tensor): Inputs to the model
    
    Returns:
      Tensor of the result
    '''
    return self.forward(inputs)
  
  def eval(self, no_track=True):
    '''Invokes EvalMode ContextManager

    Args:
      no_track (bool): If Tensors shouldn't be tracked, Defaults to False
    
    Returns:
      EvalMode ContextManager
    '''
    return EvalMode(self, no_track)
  
  def get_layers(self):
    '''Gathers all the layers in the Model

    Accomplishes by going through all its attributes and if their values are
    instances of Container/Layer it is taken as a layer

    Returns:
      Dict with attributes as key and their objects as value
    '''
    layers = {}
    for attr, val in self.__dict__.items():
      if isinstance(val, (Container, Layer)):
        layers[attr] = val
    return layers
  
  def parameters(self, as_dict=False):
    '''Gathers the params of the whole Model

    Accomplishes this by iterating through all layers and getting their params
    
    Args:
      as_dict (bool): Whether to return the params as a dict. Defaults to False
    '''
    params = {}
    for attr, layer in self.get_layers().items():
      params[attr] = layer.parameters(as_dict)
    return params if as_dict else list(list_flattener(*params.values()))
  
  def set_eval(self, eval):
    '''Sets eval

    Sets the eval of its layers to eval argument

    Args:
      eval (bool): Whether in eval mode or not
    '''
    for layer in self.get_layers().values():
      layer.set_eval(eval)
  
  def save(self, fpath):
    '''Saves the params of the model in the specified file path

    Args:
      fpath (str): File path
    '''
    params = self.parameters(as_dict=True)
    hkl.dump(params, fpath, mode='w')
    print(f"\nPARAMS SAVED at {fpath}\n")

  def load(self, fpath):
    '''Loads the params from the filepath onto the model

    Args:
      fpath (str): File path
    '''
    params = hkl.load(fpath)
    for attr, param in params.items():
      layer = self.__getattribute__(attr)
      layer.set_params(param)
    print(f"\nPARAMS LOADED from {fpath}\n")
  
  def __setattr__(self, attr, val):
    '''Sets the attributes

    Doesn't allow redefining an attribute that's been previously defined if the
    value of the attribute is instance of Container/Layer

    Args:
      attr (str): Attribute to be set
      val (object): Value to be set
    '''
    if isinstance(val, (Container, Layer)) and (attr in self.__dict__):
      raise AttributeError(f"Attribute {attr} has already been defined, it cannot be defined again for a Container/Layer")
    object.__setattr__(self, attr, val)
  
  def __repr__(self):
    return f'Model( {[str(layer) for layer in self.get_layers().values()]} )'
  
  def __str__(self):
    return f'Model( {[str(layer) for layer in self.get_layers().values()]} )'


class EvalMode:
  '''ContextManager for handling eval

  A ContextManager to run the model in eval mode, ie while testing the model.
  Use of this is that some layers like Dropout need to be turned off while
  testing
    
  Args:
    model (Model): Model to be put into eval
    no_track (bool): If True, then the backward graph is not created and the tensors
      aren't tracked
    graph (Graph): Graph that's currently in use
  '''
  def __init__(self, model, no_track):
    self.model = model
    self.no_track = no_track
    self.graph = get_graph()

  def __enter__(self):
    '''
    If no_track, then sets graph track to False
    Puts model in eval mode, by setting it to True
    '''
    if self.no_track:
      self.graph.track = False
    self.model.set_eval(True)
  
  def __exit__(self, exc_type, exc_value, exc_traceback):
    '''
    If no_track, then sets back graph track to True
    Puts model out of eval mode, by setting it to False
    '''
    if self.no_track:
      self.graph.track = True
    self.model.set_eval(False)


from .model import Model
from .layers import Sequential, Dropout, Linear, Conv2D, Conv3D, MaxPool2D, MaxPool3D
from .activations import ReLU, Sigmoid, Tanh, Softmax, LeakyReLU
from .checkpoint import Checkpoint
from .utils import save_model, load_model


from .node import Node


class Graph:
  '''Used to keep track of nodes and tensors

  The graph is constructed during the forward pass, and used by the backward
  pass to calculate gradients through automatic differentiation

  Parameters:
    graph (Graph or None): Graph object that's currently in use. If None, then the global
      _NG_GRAPH is used, else a specific graph object is used. Defaults to None
    nodes_dict (dict): Stores key-value pairs of tensors and their corresponding
      nodes in the graph
    track (bool): Whether the graph must track the tensor operations or not, ie if True, when any
      operation happens and a new result tensor is created, then the operands of the operation
      are added as parents to the result tensor and the result tensor is added as child to the
      operands, if False, none of these happens. Defaults to True
  '''

  graph = None
  
  def __init__(self):
    '''Initializes the nodes_dict to empty dict, track to True
    '''
    self.nodes_dict = {}
    self.track = True
  
  def add_edge(self, result_node, operands):
    '''Creates an edge between two nodes

    Adds edges between the result_node, which is created during an Operation, and the
    operands that produced the result. This means the result_node is added as a child of
    each of the operands and the result_node adds all operands as its parents

    Args:
      result_node (Node): node that is created in Operation.get_result_tensor
      operands (list of Tensor): All the operands for an Operation
    '''
    self.add_node(result_node)
    for operand in operands:
      if self.get_node(operand) is None:
        self.add_tensor(operand)
      operand_node = self.get_node(operand)
      result_node.add_parent(operand_node)
      operand_node.add_child(result_node)
  
  def add_node(self, node):
    '''Adds a Node to the graph

    Creates an key-value pair in nodes_dict with the specified node as the value
    and its tens attribute as the key

    Args:
      node (Node): Node to be added to the graph
    '''
    self.nodes_dict[node.tens] = node

  def get_node(self, tens):
    '''Returns the Node corresponding to the Tensor

    Args:
      tens (Tensor): Tensor whose node is to be fetched
    
    Returns:
      Node if found, else None
    '''
    return self.nodes_dict.get(tens)
  
  def add_tensor(self, tens):
    '''Adds a Tensor to the graph
    
    A new node is created for the Tensor and corresponding entry is made
    in nodes_dict

    Args:
      tens (Tensor): Tensor to be added
    '''
    self.nodes_dict[tens] = Node(tens)
  
  def remove_tensor(self, tens):
    '''Removes a Tensor from the graph

    Pops the Tensor from nodes_dict

    Args:
      tens (Tensor): Tensor to be removed
    '''
    self.nodes_dict.pop(tens)
  
  def reset_visited(self):
    '''Sets visited=False for each Node in the graph
    '''
    for node in self.nodes_dict.values():
      node.visited = False
  
  def reset_graph(self):
    '''Resets the whole graph

    This is accomplished by setting nodes_dict to an empty dictionary
    Doing so, removes all the Tensors and their Nodes from the graph
    '''
    self.nodes_dict = {}
  
  def zero_grad(self):
    '''Performs zero_grad on all the tensors in the graph

    Iterates through nodes_dict and performs zero_grad on the tensors
    '''
    for tensor in self.nodes_dict.keys():
      tensor.zero_grad()
  
  def __repr__(self):
    return 'Graph()'
  
  def __str__(self):
    return f'Graph( {self.nodes_dict} )'


import numpy as np
from ..autograd import sum as _sum, log
from ..autograd.ops.operation import Operation
from .activations import Softmax


class Loss:
  '''Base class of all loss functions
  '''
  def __call__(self, outputs, targets):
    '''Abstracts the forward method

    Args:
      outputs (Tensor): Outputs of Layer/Container/Model/Operation
      targets (Tensor): Targets to be evaluated against
    
    Returns:
      Tensor of the result
    '''
    return self.forward(outputs, targets)
  
  def get_num_examples(self, outputs_shape):
    '''Gathers the number of examples

    If dimensions of outputs_shape is 0, ie it is a scalar, then num_examples is 1
    Else the first dimension value of outputs_shape is taken as num_examples

    Args:
      outputs_shape (tuple of int): Shape of the outputs
    
    Returns:
      Number of examples
    '''
    if len(outputs_shape) in (0, 1):
      return 1
    else:
      return outputs_shape[0] 


# <------------MEANSQUAREDERROR------------>
class MSE(Loss):
  '''Mean Squared Error
  '''
  def forward(self, outputs, targets):
    '''Forward pass of MSE

    Args:
      outputs (Tensor): Outputs of Layer/Container/Model/Operation
      targets (Tensor): Targets to be evaluated against
    
    Returns:
      Tensor of the result
    '''
    num_examples = self.get_num_examples(outputs.shape)
    cost = (1/(2*num_examples))*_sum((outputs-targets)**2)
    return cost
  
  def __repr__(self):
    return f'MSE()'
  
  def __str__(self):
    return 'MeanSquaredError'


# <------------BINARYCROSSENTROPY------------>
class BCE(Loss):
  '''Binary Cross Entropy
  '''
  def forward(self, outputs, targets, epsilon=1e-9):
    '''Forward pass of BCE

    epsilon used  to prevent log0

    Args:
      outputs (Tensor): Outputs of Layer/Container/Model/Operation
      targets (Tensor): Targets to be evaluated against
      epsilon (float): For numerical stability of log Defaults to 1e-9
    
    Returns:
      Tensor of the result
    '''
    num_examples = self.get_num_examples(outputs.shape)
    entropy = _sum((outputs*log(targets+epsilon)) + ((1-outputs)*(log(1-targets+epsilon))))
    cost = (-1/num_examples)*entropy
    return cost
  
  def __repr__(self):
    return f'BCE()'
  
  def __str__(self):
    return 'BinaryCrossEntropy'


# <------------CROSSENTROPY------------>
class CE(Loss):
  '''Cross Entropy
  '''
  def forward(self, outputs, targets, epsilon=1e-9):
    '''Forward pass of CE

    epsilon used  to prevent log0

    Args:
      outputs (Tensor): Outputs of Layer/Container/Model/Operation
      targets (Tensor): Targets to be evaluated against
      epsilon (float): For numerical stability of log Defaults to 1e-9
    
    Returns:
      Tensor of the result
    '''
    num_examples = self.get_num_examples(outputs.shape)
    entropy = _sum(targets*log(outputs+epsilon))
    cost = (-1/num_examples)*entropy
    return cost
  
  def __repr__(self):
    return 'CE()'
  
  def __str__(self):
    return 'CrossEntropy'


# <------------SOFTMAXCROSSENTROPY------------>
class SoftmaxCE(Operation, Loss):
  '''Implements Softmax activation with CrossEntropyLoss

  Purpose of this is to eliminate costly Jacobian calculation involved
  with vanilla softmax activation. Since Softmax is most commonly used with
  Cross Entropy loss, if both are combined in one single Operation, then the derivative
  is a very minimal subtraction between the softmax output and the targets.
  So many intermediate backward calculations can be prevented with this.

  Parameters:
    axis (int or tuple of int): Axis along which to calculate the Softmax
      Defaults to None

  epsilon to prevent log0
  '''
  def __init__(self, axis):
    self.axis = axis

  def forward(self, outputs, targets, epsilon=1e-9):
    '''Calculates Softmax of inputs and the Cross Entropy loss

    Args:
      outputs (Tensor): Outputs of Layer/Container/Model/Operation
      targets (Tensor): Targets to be evaluated against
      epsilon (float): For numerical stability of log Defaults to 1e-9
    
    Returns:
      Tensor of the result
    '''
    num_examples = self.get_num_examples(outputs.shape)
    probs = Softmax.calc_softmax(outputs.data, axis=self.axis)
    entropy = np.sum(targets.data*np.log(probs+epsilon))
    cost = (-1/num_examples)*entropy
    return self.get_result_tensor(cost, outputs, targets)
  
  def backward(self, outputs, targets):
    '''Sets the grad_fn of outputs

    Args:
      outputs (Tensor): Tensor which is usually the outputs of the last layer
        of the network
      targets (Tensor): Targets to be evaluated against
    '''
    def sce_backward(ug):
      num_examples = self.get_num_examples(outputs.shape)
      probs = Softmax.calc_softmax(outputs.data, axis=self.axis)
      return (ug/num_examples)*(probs-targets.data) # ug is a scalar(1 by default), because loss calculated in forward is a scalar
    outputs.set_grad_fn(sce_backward)
    assert targets.requires_grad is False, 'Targets Tensor should have requires_grad=False'



import math
import numpy as np
from .operation import Operation


class Conv:
  '''Base class for Convolution and Pooling operations

  Parameters:
    padding (int): Padding value to be applied
    stride (int): Stride to be taken
  '''
  def generate_fragments(self, padded_data, kernel_shape):
    '''Generates fragments of data

    Takes the data and slices it to the shape of kernel_shape in x and y axis (the last two
    dims), including all the other dimensions, effectively taking a chunk/fragment of the data.

    Each time a chunk is taken, the x index and y index are incremented by the stride

    Args:
      padded_data (np.ndarray): Data that is already padded, to be convolved on
      kernel_shape (tuple): Shape of kernel to be convolved with
    
    Yields:
      sliced inputs_data(fragment), row slice(start and end indices of fragment among rows)
      and column slice(start and end indices of fragment among columns)

    Raises:
      AssertionError: if stride isn't greater than or equal to 1
      AssertionError: if padding isn't greater than or equal to 0
    '''
    assert self.stride>=1, 'Stride must be greater than or equal to 1'
    assert self.padding>=0, 'Padding must be greater than or equal to zero'
    padded_data_slice = ()
    padded_data_slice += ((slice(None)),)*(len(padded_data.shape)-2)
    inputs_x_dim, inputs_y_dim = padded_data.shape[-2:]
    kernel_x_dim, kernel_y_dim = kernel_shape[-2:]
    j = 0
    while(j+kernel_y_dim<=inputs_y_dim):
      i = 0
      while(i+kernel_x_dim<=inputs_x_dim):
        row_slice = slice(i, i+kernel_x_dim)
        col_slice = slice(j, j+kernel_y_dim)
        yield padded_data[padded_data_slice+(row_slice, col_slice)], row_slice, col_slice
        i+=self.stride
      j+=self.stride
  
  def get_result_shape(self, inputs_shape, kernel_shape):
    '''Calculates the x and y dimensions of result of convolution

    Takes the inputs_shape and kernel_shape and returns the x and y dims of the result
    of convolving inputs with the kernel as a 2D convolution

    Args:
      inputs_shape (tuple): Shape of inputs
      kernel_shape (tuple): Shape of kernel
    
    Returns:
      tuple of x and y dims of result
    '''
    inputs_x_dim, inputs_y_dim = inputs_shape[-2:]
    kernel_x_dim, kernel_y_dim = kernel_shape[-2:]
    def result_dim(inputs_dim, kernel_dim):
      return math.floor(((inputs_dim + (2*self.padding) - kernel_dim)/self.stride) + 1)
    result_x_dim = result_dim(inputs_x_dim, kernel_x_dim)
    result_y_dim = result_dim(inputs_y_dim, kernel_y_dim)
    return result_x_dim, result_y_dim
  
  def pad(self, data):
    '''Pads the data in x and y dimensions (last two dims)

    Only the last two dimensions are padded rest aren't padded(padded with 0, 
    has no effect)

    Args:
      data (np.ndarray): Data to be padded
    
    Returns:
      data that is padded
    '''
    padding = () 
    padding+=((0,0),)*(len(data.shape)-2)
    padding+=((self.padding,self.padding),)*2
    return np.pad(data, padding)

  def unpad(self, padded_data):
    '''Unpads the padded data

    Slices the padded_data in last two dimensions where it is padded, by rejecting
    the padding and only extracting the original data

    Args:
      padded_data (np.ndarray): Data that needs to be unpadded
    
    Returns:
      Unpadded data
    '''
    padded_x_dim, padded_y_dim = padded_data.shape[-2:]
    extractor_slice = ()
    extractor_slice+=((slice(None)),)*(len(padded_data.shape)-2)
    extractor_slice+=(slice(self.padding,padded_x_dim-self.padding), slice(self.padding,padded_y_dim-self.padding))
    return padded_data[extractor_slice]
  
  def fragment_iterator(self, padded_inputs, kernel_shape, *args):
    '''Zips the fragments with any other args

    Args:
      padded_inputs (np.ndarray): Inputs that are padded
      kernel_shape (tuple): Shape of the kernel
      *args: Any other objects that should be zipped together, this is usually the
        upper gradient chunks that are the result of the convolution
    
    Returns:
      zip of fragments and ony other objects in args
    '''
    return zip(self.generate_fragments(padded_inputs, kernel_shape), *args)


# <------------CONV2D------------>
class Conv2D(Operation, Conv):
  '''Implements 2D convolution

  2D convolution where the inputs has only 1 channel and its shape is of the form
  (num_examples, x_dim, y_dim) is convolved with a 2D kernel
  '''
  def __init__(self, padding, stride):
    self.padding = padding
    self.stride = stride
  
  def forward(self, inputs, kernel, bias):
    '''Implements the forward pass

    Each fragment of shape (num_examples, kernel_shape[0], kernel_shape[1]) is element wise
    multipled with the kernel and then summed along its x and y axis and then adds it with the bias

    Args:
      inputs (Tensor or int or float or list or np.ndarray): Tensor to be convolved on
      kernel (Tensor or int or float or list or np.ndarray): Tensor to be convolved with(weights)
      bias (Tensor or int or float or list or np.ndarray): bias value
    
    Returns:
      Tensor of the result that is convolved
    '''
    inputs, kernel, bias = self.get_tensors(inputs, kernel, bias)
    self.validate_inputs(inputs)
    outputs = np.empty((inputs.shape[0], *self.get_result_shape(inputs.shape, kernel.shape)))
    padded_inputs = self.pad(inputs.data)
    for (fragment, _, _), idx in self.fragment_iterator(padded_inputs, kernel.shape, np.ndindex(outputs.shape[-2:])):
      output = np.sum((fragment*kernel.data), axis=(1,2)) + bias.data
      outputs[:,idx[0],idx[1]] = output
    return self.get_result_tensor(outputs, inputs, kernel, bias)
  
  def backward(self, inputs, kernel, bias):
    '''Sets the grad_fn of inputs, kernel and bias

    Since each convolution of a fragment with kernel results in a Tensor, the corresponding upper
    gradient values of all the examples are taken.
    
    inputs_grads are initialized to zero of shape of padded_inputs, then the corresponding gradient
    that is calculated is tucked into the inputs_grads based on the row slice and the column slice.
    The inputs_grads are then unpadded to reject the gradients of pads and only keeps the gradients
    of the original inputs

    kernel_grads are intitialized to zero and since they are used multiple times, its gradient is added
    each time

    Since bias is added to all outputs, its gradient is just
    the sum of the upper gradient

    Args:
      inputs (Tensor): Tensor that is convolved on
      kernel (Tensor): Tensor that is convolved with(weights)
      bias (Tensor): bias value
    '''
    from ..utils import unbroadcast_data
    padded_inputs = self.pad(inputs.data)

    def inputs_backward(ug):
      inputs_grads = np.zeros(padded_inputs.shape)
      for (fragment, row_slice, col_slice), idx in self.fragment_iterator(padded_inputs, kernel.shape, np.ndindex(ug.shape[-2:])):
        sliced_ug = ug[:,idx[0],idx[1]]
        sum_grad = np.ones(fragment.shape)*sliced_ug.reshape(sliced_ug.size,1,1)
        fragment_grad = kernel.data*sum_grad
        inputs_grads[:, row_slice, col_slice]+=fragment_grad
      unpadded_inputs_grads = self.unpad(inputs_grads)
      return unpadded_inputs_grads

    def kernel_backward(ug):
      kernel_grads = np.zeros(kernel.shape)
      for (fragment, _, _), idx in self.fragment_iterator(padded_inputs, kernel.shape, np.ndindex(ug.shape[-2:])):
        sliced_ug = ug[:,idx[0],idx[1]]
        sum_grad = np.ones(fragment.shape)*sliced_ug.reshape(sliced_ug.size,1,1)
        kernel_grad = unbroadcast_data(fragment*sum_grad, kernel.shape, fragment.shape)
        kernel_grads+=kernel_grad
      return kernel_grads

    def bias_backward(ug):
      return np.sum(ug)
      
    inputs.set_grad_fn(inputs_backward)
    kernel.set_grad_fn(kernel_backward)
    bias.set_grad_fn(bias_backward)
  
  def validate_inputs(self, inputs):
    '''Validates the inputs

    Args:
      inputs (Tensor or np.ndarray): Data to be validated

    Raises:
      ValueError: If inputs aren't of format (num_examples, x_dim, y_dim)
    '''
    if len(inputs.shape)!=3: # The first dimension should be number of examples
      raise ValueError("Only 3D inputs, with 0th dim as number of examples are supported!")

def conv2d(inputs, kernel, bias, padding, stride):
  '''Abstraction of Conv2D.forward

  Args:
    inputs (Tensor or int or float or list or np.ndarray): Tensor to be convolved on
    kernel (Tensor or int or float or list or np.ndarray): Tensor to be convolved with(weights)
    bias (Tensor or int or float or list or np.ndarray): bias value
    padding (int): Padding value to be applied
    stride (int): Stride to be taken

  Returns:
    Tensor of the result
  '''
  return Conv2D(padding, stride).forward(inputs, kernel, bias)


# <------------CONV3D------------>
class Conv3D(Operation, Conv):
  '''Implements 3D convolution

  3D convolution over a colume where the inputs has multiple channels and
  its shape is of the form (num_examples, num_channels, x_dim, y_dim) is convolved
  with a 3D kernel of shape (num_channels, kernel_shape[0], kernel_shape[1])
  '''
  def __init__(self, padding, stride):
    self.padding = padding
    self.stride = stride
  
  def forward(self, inputs, kernel, bias):
    '''Implements the forward pass

    Each fragment of shape (num_examples, num_channels, kernel_shape[0], kernel_shape[1])
    is element wise multipled with the kernel and then summed along its x, y and z axis and
    then adds it with the bias

    Args:
      inputs (Tensor or int or float or list or np.ndarray): Tensor to be convolved on
      kernel (Tensor or int or float or list or np.ndarray): Tensor to be convolved with(weights)
      bias (Tensor or int or float or list or np.ndarray): bias value
    
    Returns:
      Tensor of the result that is convolved
    '''
    inputs, kernel, bias = self.get_tensors(inputs, kernel, bias)
    self.validate_inputs(inputs)
    outputs = np.empty((inputs.shape[0], kernel.shape[0], *self.get_result_shape(inputs.shape, kernel.shape)))
    padded_inputs = self.pad(inputs.data)
    for (fragment,_,_), idx in self.fragment_iterator(padded_inputs, kernel.shape, np.ndindex(outputs.shape[-2:])):
      expanded_fragment = np.expand_dims(fragment, axis=1)
      output = expanded_fragment*kernel.data
      output = np.sum(output, axis=(2,3,4)) + bias.data
      outputs[:,:,idx[0],idx[1]] = output
    return self.get_result_tensor(outputs, inputs, kernel, bias)
  
  def backward(self, inputs, kernel, bias):
    '''Sets the grad_fn of inputs, kernel and bias

    Since each convolution of a fragment with kernel results in a Tensor, the corresponding upper
    gradient values of all the examples, across all channels are taken.
    
    To calculate sum_grad which is the gradient of the sum operation during forward pass, the fragment
    needs to be expanded along first axis to allow for broadcasting of upper gradient slice.

    inputs_grads are initialized to zero of shape of padded_inputs, then the corresponding gradient
    that is calculated is tucked into the inputs_grads based on the row slice and the column slice.
    The inputs_grads are then unpadded to reject the gradients of pads and only keeps the gradients
    of the original inputs

    kernel_grads are intitialized to zero and since they are used multiple times, its gradient is added
    each time

    Since bias is a vector here and is not added to all the outputs, it is only summed across all the examples
    and the first and second axis

    Args:
      inputs (Tensor): Tensor that is convolved on
      kernel (Tensor): Tensor that is convolved with(weights)
      bias (Tensor): bias value
    '''
    from ..utils import unbroadcast_data
    padded_inputs = self.pad(inputs.data)

    def inputs_backward(ug):
      inputs_grads = np.zeros(padded_inputs.shape)
      for (fragment, row_slice, col_slice), idx in self.fragment_iterator(padded_inputs, kernel.shape, np.ndindex(ug.shape[-2:])):
        expanded_fragment = np.expand_dims(fragment, axis=1)
        sliced_ug = ug[:,:,idx[0],idx[1]]
        sliced_ug = sliced_ug.reshape(*sliced_ug.shape,1,1,1)
        sum_grad = np.ones(expanded_fragment.shape)*sliced_ug
        fragment_grad = np.sum(sum_grad*kernel.data, axis=1)
        inputs_grads[:,:,row_slice,col_slice]+=fragment_grad
      unpadded_inputs_grads = self.unpad(inputs_grads)
      return unpadded_inputs_grads
    
    def kernel_backward(ug):
      kernel_grads = np.zeros(kernel.shape)
      for (fragment,_,_), idx in self.fragment_iterator(padded_inputs, kernel.shape, np.ndindex(ug.shape[-2:])):
        expanded_fragment = np.expand_dims(fragment,1)
        sliced_ug = ug[:,:,idx[0],idx[1]]
        sliced_ug = sliced_ug.reshape(*sliced_ug.shape,1,1,1)
        sum_grad = np.ones(expanded_fragment.shape)*sliced_ug
        kernel_grad = sum_grad*expanded_fragment
        kernel_grad = unbroadcast_data(kernel_grad, kernel.shape, kernel_grad.shape)
        kernel_grads+=kernel_grad
      return kernel_grads
    
    def bias_backward(ug):
      grad = np.sum(ug, axis=0)
      grad = np.sum(grad, axis=2, keepdims=True)
      grad = np.sum(grad, axis=1, keepdims=True)
      return grad
    
    inputs.set_grad_fn(inputs_backward)
    kernel.set_grad_fn(kernel_backward)
    bias.set_grad_fn(bias_backward)
  
  def validate_inputs(self, inputs):
    '''Validates the inputs

    Args:
      inputs (Tensor or np.ndarray): Data to be validated

    Raises:
      ValueError: If inputs aren't of format (num_examples, num_channels, x_dim, y_dim)
    '''
    if len(inputs.shape)!=4:
      raise ValueError("Only 4D inputs, with 0th dim as number of examples, 1st dim as number of channels are supported!")

def conv3d(inputs, kernel, bias, padding, stride):
  '''Abstraction of Conv3D.forward

  Args:
    inputs (Tensor or int or float or list or np.ndarray): Tensor to be convolved on
    kernel (Tensor or int or float or list or np.ndarray): Tensor to be convolved with(weights)
    bias (Tensor or int or float or list or np.ndarray): bias value
    padding (int): Padding value to be applied
    stride (int): Stride to be taken

  Returns:
    Tensor of the result
  '''
  return Conv3D(padding, stride).forward(inputs, kernel, bias)


# <------------MAXPOOL2D------------>
class MaxPool2D(Operation, Conv):
  '''Implements 2D MaxPooling

  In MaxPooling, it is differentiable only if kernel_shape along with the stride
  covers the entire input if not it is not differentiable and fails gradient
  checking

  Parameters:
    kernel_shape (tuple): Shape of the kernel
  '''
  def __init__(self, kernel_shape, padding, stride):
    '''
    Raises:
      ValueError: if kernel shape dimensions aren't 2D
    '''
    if len(kernel_shape)==2:
      self.kernel_shape = kernel_shape
    else:
      raise ValueError("Only 2D kernels are allowed!")
    self.padding = padding
    self.stride = stride
  
  def forward(self, inputs):
    '''Forward pass of max pooling 2d

    The fragments are generated, for each of it, the maximum value in the x and y dims
    is returned for all examples

    Args:
      inputs (Tensor or int or float or list or np.ndarray): Data to be maxpooled
    
    Returns:
      Tensor of the result
    '''
    inputs = self.get_tensors(inputs)
    self.validate_inputs(inputs)
    outputs = np.empty((inputs.shape[0], *self.get_result_shape(inputs.shape, self.kernel_shape)))
    padded_inputs = self.pad(inputs.data)
    for (fragment,_,_), idx in self.fragment_iterator(padded_inputs, self.kernel_shape, np.ndindex(outputs.shape[-2:])):
      outputs[:,idx[0],idx[1]] = np.max(fragment, axis=(1,2))
    return self.get_result_tensor(outputs, inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of inputs

    Since argmax operates only on one axis, the fragment is first flattened across x and
    y dims (last two dims)
    Then the one hot encoding of the max indices are taken that are the multiplied
    with the corresponding upper grad slice
    fragment_grad is reshaped to original fragment shape

    Args:
      inputs (Tensor): Tensor that is maxpooled
    '''
    padded_inputs = self.pad(inputs.data)

    def inputs_backward(ug):
      inputs_grad = np.empty(padded_inputs.shape)
      for (fragment,row_slice,col_slice),idx in self.fragment_iterator(padded_inputs, self.kernel_shape, np.ndindex(ug.shape[-2:])):
        sliced_ug = ug[:,idx[0],idx[1]]
        fragment_shape = fragment.shape
        flattened_fragment = fragment.reshape(fragment_shape[0], fragment_shape[1]*fragment_shape[2])
        args = np.argmax(flattened_fragment, axis=-1)
        fragment_grad = np.eye(flattened_fragment.shape[-1])[args] # one hot encoding of args
        fragment_grad = fragment_grad.reshape(fragment_shape)
        inputs_grad[:,row_slice,col_slice] = fragment_grad*np.expand_dims(sliced_ug,axis=(1,2))
      unpadded_inputs_grads = self.unpad(inputs_grad)
      return unpadded_inputs_grads

    inputs.set_grad_fn(inputs_backward)
  
  def validate_inputs(self, inputs):
    '''Validates the inputs

    Args:
      inputs (Tensor or np.ndarray): Data to be validated

    Raises:
      ValueError: If inputs aren't of format (num_examples, x_dim, y_dim)
    '''
    if len(inputs.shape)!=3: # The first dimension should be number of examples
      raise ValueError("Only 3D inputs, with 0th dim as number of examples are supported!")

def maxpool2d(inputs, kernel_shape, padding, stride):
  '''Abstraction of MaxPool2D.forward

  Args:
    inputs (Tensor or int or float or list or np.ndarray): Tensor to be convolved on
    kernel_shape (tuple): Shape of the kernel
    padding (int): Padding value to be applied
    stride (int): Stride to be taken

  Returns:
    Tensor of the result
  '''
  return MaxPool2D(kernel_shape, padding, stride).forward(inputs)


# <------------MAXPOOL3D------------>
class MaxPool3D(Operation, Conv):
  '''Implements 3D MaxPooling

  In MaxPooling, it is differentiable only if kernel_shape along with the stride
  covers the entire input if not it is not differentiable and fails gradient
  checking

  Parameters:
    kernel_shape (tuple): Shape of the kernel
  '''
  def __init__(self, kernel_shape, padding, stride):
    '''
    Raises:
      ValueError: if kernel shape dimensions aren't 2D
    '''
    if len(kernel_shape)==2:
      self.kernel_shape = kernel_shape
    else:
      raise ValueError("Only 2D kernels are allowed!")
    self.padding = padding
    self.stride = stride
  
  def forward(self, inputs):
    '''Forward pass of max pooling 3d

    The fragments are generated, for each of it, the maximum value in the x and y dims
    is returned for all examples across all channels

    Args:
      inputs (Tensor or int or float or list or np.ndarray): Data to be maxpooled
    
    Returns:
      Tensor of the result
    '''
    inputs = self.get_tensors(inputs)
    self.validate_inputs(inputs)
    outputs = np.empty((inputs.shape[0], inputs.shape[1], *self.get_result_shape(inputs.shape, self.kernel_shape)))
    padded_inputs = self.pad(inputs.data)
    for (fragment,_,_), idx in self.fragment_iterator(padded_inputs, self.kernel_shape, np.ndindex(outputs.shape[-2:])):
      outputs[:,:,idx[0],idx[1]] = np.max(fragment, axis=(2,3))
    return self.get_result_tensor(outputs, inputs)
  
  def backward(self, inputs):
    '''Sets the grad_fn of inputs

    Since argmax operates only on one axis, the fragment is first flattened across x and
    y dims (last two dims)
    Then the one hot encoding of the max indices are taken that are the multiplied
    with the corresponding upper grad slice
    fragment_grad is reshaped to original fragment shape

    Args:
      inputs (Tensor): Tensor that is maxpooled
    '''
    padded_inputs = self.pad(inputs.data)

    def inputs_backward(ug):
      inputs_grad = np.empty(padded_inputs.shape)
      for (fragment,row_slice,col_slice),idx in self.fragment_iterator(padded_inputs, self.kernel_shape, np.ndindex(ug.shape[-2:])):
        sliced_ug = ug[:,:,idx[0],idx[1]]
        fragment_shape = fragment.shape
        flattened_fragment = fragment.reshape(fragment_shape[0]*fragment_shape[1], fragment_shape[2]*fragment_shape[3])
        args = np.argmax(flattened_fragment, axis=-1)
        fragment_grad = np.eye(flattened_fragment.shape[-1])[args] # one hot encoding of args
        fragment_grad = fragment_grad.reshape(fragment_shape)
        inputs_grad[:,:,row_slice,col_slice] = fragment_grad*np.expand_dims(sliced_ug,axis=(2,3))
      unpadded_inputs_grads = self.unpad(inputs_grad)
      return unpadded_inputs_grads
    
    inputs.set_grad_fn(inputs_backward)
  
  def validate_inputs(self, inputs):
    '''Validates the inputs

    Args:
      inputs (Tensor or np.ndarray): Data to be validated

    Raises:
      ValueError: If inputs aren't of format (num_examples, num_channels, x_dim, y_dim)
    '''
    if len(inputs.shape)!=4:
      raise ValueError("Only 4D inputs, with 0th dim as number of examples, 1st dim as number of channels are supported!")

def maxpool3d(inputs, kernel_shape, padding, stride):
  '''Abstraction of MaxPool3D.forward

  Args:
    inputs (Tensor or int or float or list or np.ndarray): Tensor to be convolved on
    kernel_shape (tuple): Shape of the kernel
    padding (int): Padding value to be applied
    stride (int): Stride to be taken

  Returns:
    Tensor of the result
  '''
  return MaxPool3D(kernel_shape, padding, stride).forward(inputs)


import numpy as np
from ..node import Node


class Operation:
  '''Transforms Tensors by applying some function

  Used when some input is getting transformed into an output, for functions
  where gradient calculation is required with the forward pass and the backward
  pass defined
  '''
  
  def process_operands(self, operands):
    '''All operands are converted to Tensors

    Args:
      operands (Tensor or int or float or list or np.ndarray): Operands of the Operation
    
    Returns:
      tuple of Tensors
    '''
    from ..tensor import Tensor
    operands = list(operands)
    for i,operand in enumerate(operands):
      if not isinstance(operand, Tensor):
        operands[i] = Tensor(operand)
    return tuple(operands)
  
  def get_tensors(self, *operands):
    '''Returns the processed operands as tuple of Tensors

    Args:
      *operands (Tensor or int or float or list or np.ndarray): Operands of the Operation
    
    Returns:
      tuple of Tensors if len(tuple)>1 else returns the first Tensor
    '''
    tensors = self.process_operands(operands)
    if len(tensors)==0:
      return None
    elif len(tensors)==1:
      return tensors[0]
    else:
      return tensors
  
  def get_broadcast_shape(self, *tensors):
    '''Return broadcasted shape of Tensors

    If the tensors can be broadcasted, then the broadcasted shape is returned
    , else None.

    Args:
      *tensors (Tensor): Tensors that should be broadcasted

    Returns:
      Broadcasted shape if it can be broadcasted, if not None
      Also even if atleast one of the Tensors has requires_broadcasting set to False,
      it returns None
    '''
    for tens in tensors:
      if not(tens.requires_broadcasting):
        return None
    try:
      return np.broadcast_shapes(*(tens.data.shape for tens in tensors))
    except ValueError:
      return None
  
  def result_requires_grad(self, tensors):
    '''Checks if the result requires grad

    Checks if the result requires gradient to be calculated given the operands of the
    Operation, if atleast one operand requires_grad to True, then result will also have
    requires_grad to True

    Args:
      tensors (Tensor): Tensors that are operated on
    '''
    for tens in tensors:
      if tens.requires_grad:
        return True
    return False
  
  def get_result_tensor(self, result, *tensors):
    '''Returns the result tensor of the Operation
    
    If tracking is enabled, then, it creates a Node for the result_tensor
    with parent_broadcast_shape and adds edges to the graph
    
    If tracking is disabled, then no Node creation and edge addition
    occurs

    Args:
      result (np object): Result after performing a raw numpy operation
      *tensors (Tensor): Operands of the operation

    Returns:
      Tensor of the result
    '''
    from ..tensor import Tensor
    from ..utils import get_graph
    graph = get_graph()
    result = result.astype(np.ndarray)
    result_tensor = Tensor(result, self.result_requires_grad(tensors))
    if graph.track:
      result_node = Node(result_tensor)
      result_node.backward_fn = self.backward
      result_node.parent_broadcast_shape = self.get_broadcast_shape(*tensors)
      graph.add_edge(result_node, tensors)
    return result_tensor
  
  def backward(self, *args):
    '''Abstract backward method

    Raises:
      NotImplementedError: If backward method isn't overridden
    '''
    raise NotImplementedError(f"Backward method not implemented for Operation {self}")


import numpy as np
from .operation import Operation


# <------------ADD------------>
class Add(Operation):
  '''Element wise addition between two Tensors or Tensor-like
  '''
  def forward(self, tens1, tens2):
    '''Calculates element wise addition

    Args:
      tens1 (Tensor or int or float or list or np.ndarray): First operand
      tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
    Returns:
      Tensor of the result
    '''
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data+tens2.data, tens1, tens2)

  def backward(self, tens1, tens2):
    '''Sets grad_fn of operands

    Local gradient is an identity matrix, that should be dotted with the upper gradient
    which results in upper gradient

    Args:
      tens1 (Tensor): First operand
      tens2 (Tensor): Second operand
    '''
    tens1.set_grad_fn(lambda ug:ug)
    tens2.set_grad_fn(lambda ug:ug)

def add(tens1, tens2):
  '''Abstraction for Add.forward

  Args:
    tens1 (Tensor or int or float or list or np.ndarray): First operand
    tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
  Returns:
    Tensor of the result
  '''
  return Add().forward(tens1, tens2)


# <------------SUB------------>
class Sub(Operation):
  '''Element wise subtraction between two Tensors or Tensor-like
  '''
  def forward(self, tens1, tens2):
    '''Calculates element wise subtraction

    Args:
      tens1 (Tensor or int or float or list or np.ndarray): First operand
      tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
    Returns:
      Tensor of the result
    '''
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data-tens2.data, tens1, tens2)
  
  def backward(self, tens1, tens2):
    '''Sets grad_fn of operands

    Local gradient is an identity matrix, that should be dotted with the upper gradient
    which results in upper gradient, for the other one local gradient is a negative identity
    matrix which results in negative upper gradient

    Args:
      tens1 (Tensor): First operand
      tens2 (Tensor): Second operand
    '''
    tens1.set_grad_fn(lambda ug:ug)
    tens2.set_grad_fn(lambda ug:-ug)

def sub(tens1, tens2):
  '''Abstraction for Sub.forward

  Args:
    tens1 (Tensor or int or float or list or np.ndarray): First operand
    tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
  Returns:
    Tensor of the result
  '''
  return Sub().forward(tens1, tens2)


# <------------MUL------------>
class Mul(Operation):
  '''Element wise multiplication between two Tensors or Tensor-like
  '''
  def forward(self, tens1, tens2):
    '''Calculates element wise multiplication

    Args:
      tens1 (Tensor or int or float or list or np.ndarray): First operand
      tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
    Returns:
      Tensor of the result
    '''
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data*tens2.data, tens1, tens2)
  
  def backward(self, tens1, tens2):
    '''Sets grad_fn of operands

    Local gradient for each Tensor is the other Tensor's data, which is element-wise
    multiplied with upper gradient

    Args:
      tens1 (Tensor): First operand
      tens2 (Tensor): Second operand
    '''
    tens1.set_grad_fn(lambda ug:tens2.data*ug)
    tens2.set_grad_fn(lambda ug:tens1.data*ug)

def mul(tens1, tens2):
  '''Abstraction for Mul.forward

  Args:
    tens1 (Tensor or int or float or list or np.ndarray): First operand
    tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
  Returns:
    Tensor of the result
  '''
  return Mul().forward(tens1, tens2)


# <------------DIV------------>
class Div(Operation):
  '''Element wise division between two Tensors or Tensor-like
  '''
  def forward(self, tens1, tens2):
    '''Calculates element wise division

    Args:
      tens1 (Tensor or int or float or list or np.ndarray): First operand
      tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
    Returns:
      Tensor of the result
    '''
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(tens1.data/tens2.data, tens1, tens2)
  
  def backward(self, tens1, tens2):
    '''Sets grad_fn of operands

    Local gradient of tens1 is 1/tens2.data, local gradient of tens2 is
    -1*tens1.data/tens2.data^2, which is element wise multiplied with upper
    gradient

    Args:
      tens1 (Tensor): First operand
      tens2 (Tensor): Second operand
    '''
    tens1.set_grad_fn(lambda ug:(1/tens2.data)*ug)
    tens2.set_grad_fn(lambda ug:((-1*tens1.data)/np.power(tens2.data, 2))*ug)

def div(tens1, tens2):
  '''Abstraction for Div.forward

  Args:
    tens1 (Tensor or int or float or list or np.ndarray): First operand
    tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
  Returns:
    Tensor of the result
  '''
  return Div().forward(tens1, tens2)


# <------------DOT------------>
class Dot(Operation):
  '''Dot product between two Tensors or Tensor-like
  '''
  def forward(self, tens1, tens2):
    '''Calculates dot product

    Args:
      tens1 (Tensor or int or float or list or np.ndarray): First operand
      tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
    Returns:
      Tensor of the result
    '''
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(np.dot(tens1.data, tens2.data), tens1, tens2)
  
  def backward(self, tens1, tens2):
    '''Sets grad_fn of operands

    Local gradient of tens1 is transpose of tens2.data, local gradient of tens2 is
    transpose of tens1.data, which is dotted with upper gradient

    Args:
      tens1 (Tensor): First operand
      tens2 (Tensor): Second operand
    '''
    tens1.set_grad_fn(lambda ug:np.dot(ug, tens2.data.T))
    tens2.set_grad_fn(lambda ug:np.dot(tens1.data.T, ug))

def dot(tens1, tens2):
  '''Abstraction for Dot.forward

  Args:
    tens1 (Tensor or int or float or list or np.ndarray): First operand
    tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
  Returns:
    Tensor of the result
  '''
  return Dot().forward(tens1, tens2)


# <------------EXP------------>
class Exp(Operation):
  '''Exponentiates the Tensor or Tensor-like
  '''
  def forward(self, tens):
    '''Calculates exponentiation

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    
    Returns:
      Tensor of the result
    '''
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.exp(tens.data), tens)
  
  def backward(self, tens):
    '''Sets grad_fn of operand

    Local gradient is exponentiation of tens.data itself

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    '''
    tens.set_grad_fn(lambda ug:np.exp(tens.data)*ug)

def exp(tens):
  '''Abstraction for Exp.forward

  Args:
    tens (Tensor): Operand
  
  Returns:
    Tensor of the result
  '''
  return Exp().forward(tens)


# <------------LOG------------>
class Log(Operation):
  '''Natural Logarithm of the Tensor or Tensor-like
  '''
  def forward(self, tens):
    '''Calculates natural logarithm

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    
    Returns:
      Tensor of the result
    '''
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.log(tens.data), tens)
  
  def backward(self, tens):
    '''Sets grad_fn of operand

    Local gradient is exponentiation of 1/tens.data itself

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    '''
    tens.set_grad_fn(lambda ug:(1/tens.data)*ug)

def log(tens):
  '''Abstraction for Log.forward

  Args:
    tens (Tensor): Operand
  
  Returns:
    Tensor of the result
  '''
  return Log().forward(tens)


# <------------POW------------>
class Pow(Operation):
  '''Raises one Tensor or Tensor-like to the power of another Tensor or Tensor-like
  '''
  def forward(self, tens1, tens2):
    '''Calculates raising to a power

    Args:
      tens1 (Tensor or int or float or list or np.ndarray): First operand
      tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
    Returns:
      Tensor of the result
    '''
    tens1, tens2 = self.get_tensors(tens1, tens2)
    return self.get_result_tensor(np.power(tens1.data, tens2.data), tens1, tens2)
  
  def backward(self, tens1, tens2):
    '''Sets grad_fn of operands

    Local gradient of tens1 is tens1.data^(tens2.data-1), local gradient of tens2 is
    log(tens1.data), which is element wise multiplied with upper gradient

    Args:
      tens1 (Tensor): First operand
      tens2 (Tensor): Second operand
    '''
    result = np.power(tens1.data, tens2.data)
    tens1.set_grad_fn(lambda ug:(np.power(tens1.data, tens2.data-1) * tens2.data)*ug)
    tens2.set_grad_fn(lambda ug:(result*np.log(tens1.data))*ug)

def pow(tens1, tens2):
  '''Abstraction for Pow.forward

  Args:
    tens1 (Tensor or int or float or list or np.ndarray): First operand
    tens2 (Tensor or int or float or list or np.ndarray): Second operand
    
  Returns:
    Tensor of the result
  '''
  return Pow().forward(tens1, tens2)


# <------------SUM------------>
class Sum(Operation):
  '''Performs sum along a specified axis

  If axis is None, then the sum of the entire Tensor is calculated

  Parameters:
    axis (None or int or tuple of int): Axis along which it should be summed
  '''
  def __init__(self, axis=None):
    '''
    Args:
      axis (None or int or tuple of int): Axis along which it should be summed
        Defaults to None
    '''
    self.axis = axis
  
  def forward(self, tens):
    '''Calculates sum along an axis

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    
    Returns:
      Tensor of the result
    '''
    tens = self.get_tensors(tens)
    return self.get_result_tensor(np.sum(tens.data, axis=self.axis), tens)
  
  def backward(self, tens):
    '''Sets grad_fn of operand

    Local gradient is all ones and the upper gradient must be added a new axis
    along the axis attribute if axis is not None, for broadcasting of upper_gradient
    as during forward pass the dimension will be reduced along the axis it is summed

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    '''
    def sum_backward(ug):
      if self.axis is not None:
        ug = np.expand_dims(ug, axis=self.axis)
      grads = np.ones(tens.shape)*ug
      return grads
    tens.set_grad_fn(sum_backward)

def sum(tens, axis=None):
  '''Abstraction for Sum.forward

  Args:
    tens (Tensor): Operand
    axis (None or int or tuple of int): Axis along which it should be summed
      Defaults to None
  
  Returns:
    Tensor of the result
  '''
  return Sum(axis).forward(tens)


# <------------TRANSPOSE------------>
class Transpose(Operation):
  '''Performs transpose of Tensor or Tensor-like
  '''
  def forward(self, tens):
    '''Performs transpose

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    
    Returns:
      Tensor of the result
    '''
    tens = self.get_tensors(tens)
    return self.get_result_tensor(tens.data.T, tens)

  def backward(self, tens):
    '''Sets grad_fn of operand

    No local gradient, upper gradient is just transposed

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    '''
    tens.set_grad_fn(lambda ug:ug.T)

def transpose(tens):
  '''Abstraction for Transpose.forward

  Args:
    tens (Tensor): Operand
  
  Returns:
    Tensor of the result
  '''
  return Transpose().forward(tens)


# <------------FLATTEN------------>
class Flatten(Operation):
  '''Performs flattening of Tensor or Tensor-like
  '''
  def forward(self, tens):
    '''Performs flattening from any dimension to 1D

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    
    Returns:
      Tensor of the result
    '''
    tens = self.get_tensors(tens)
    flattened = tens.data.flatten()
    return self.get_result_tensor(flattened.reshape(flattened.shape[0],1), tens)
  
  def backward(self, tens):
    '''Sets grad_fn of operand

    No local gradient, upper gradient is reshaped to original shape

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    '''
    tens.set_grad_fn(lambda ug:ug.reshape(tens.shape))

def flatten(tens):
  '''Abstraction for Flatten.forward

  Args:
    tens (Tensor): Operand
  
  Returns:
    Tensor of the result
  '''
  return Flatten().forward(tens)


# <------------RESHAPE------------>
class Reshape(Operation):
  '''Performs reshaping of Tensor or Tensor-like
  '''
  def forward(self, tens, new_shape):
    '''Performs reshaping of Tensor to a new shape

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
      new_shape (tuple): New shape to be reshaped into
    
    Returns:
      Tensor of the result
    '''
    tens = self.get_tensors(tens)
    return self.get_result_tensor(tens.data.reshape(new_shape), tens)
  
  def backward(self, tens):
    '''Sets grad_fn of operand

    No local gradient, upper gradient is reshaped to original shape

    Args:
      tens (Tensor or int or float or list or np.ndarray): Operand
    '''
    tens.set_grad_fn(lambda ug:ug.reshape(tens.shape))

def reshape(tens, new_shape):
  '''Abstraction for Reshape.forward

  Args:
    tens (Tensor): Operand
  
  Returns:
    Tensor of the result
  '''
  return Reshape().forward(tens, new_shape)


import os
import json
import numpy as np
import datetime as dt
import secrets
from hashlib import sha256


class Checkpoint:
  '''Creates and initializes files for checkpoints
    
  A JSON file checkpoints.json is created which contains the dict which has the tracked values
  and also the params file name at the time of adding a checkpoint, operates in append mode
  A new params file is created at each checkpoint and the JSON file is updated

  Parameters:
    session (str): Current session that is in use
    dirpath (str): Directory in which checkpoints must be saved
    model (Model): Model to be checkpointed
    hash_length (int): Character length of session identifiers. Defaults to 16
  '''
  def __init__(self, model, dirpath, hash_length=16):
    '''
    Raises:
      AssertionError: if hash_length<0 and hash_length>64
    '''
    self.session = None
    self.dirpath = None
    assert hash_length>0 and hash_length<=64, 'Hash length must be between 1 and 64'
    self.hash_length = hash_length
    self._init_files(dirpath)
    self.model = model
  
  def new_session(self):
    '''Creates a new session in checkpoints.json

    Also creates a new directory for the session

    Returns:
      self
    '''
    self.session = self._generate_hash()
    os.mkdir(f'{self.dirpath}/{self.session}')
    return self
  
  def specify_session(self, session):
    '''Used to specify a particular session to use for checkpoints.json

    Args:
      session (str): The session to be used
    
    Raises:
      ValueError: If session is not already in checkpoints.json
    '''
    with open(f'{self.dirpath}/checkpoints.json', 'r') as checkpoints_fp:
      existing_checkpoints = json.load(checkpoints_fp)
      if session in existing_checkpoints:
        self.session = session
      else:
        raise ValueError(f"Invalid session {session} specified!")

  def add(self, **tracked):
    '''Adds a new checkpoint

    Args:
      **tracked: All the data that needs to be tracked in checkpoints.json
    
    Raises:
      ValueError: If forbidden_attrs ('datetime') are used as keys in tracked, because
        the same key is used by neograd to add key of the same value, which might get overwritten
      ValueError: If values in tracked aren't serializable and don't belong to builtin classes
    '''
    forbidden_attrs = ('datetime')
    allowed_types = (int, float, str, dict, list)
    for attr, val in tracked.items():
      if not(isinstance(val, allowed_types)):
        raise ValueError(f'Only {allowed_types} can be tracked, if using a Tensor, use data attribute and convert it into native python objects or strings!')
      if attr in forbidden_attrs:
        raise ValueError(f'Attribute {attr} must not be present in forbidden_attrs {forbidden_attrs}, as neograd uses them internally!')
    updated_checkpoint, params_fname_hash = self._update(tracked)
    self._save(updated_checkpoint, params_fname_hash)
  
  def _update(self, new_checkpoint):
    '''Updates the new_checkpoint

    Updates the checkpoint by including the datetime of adding new checkpoint
    Generates the hash that'll be used as the filename for the params file that'll be saved

    Args:
      new_checkpoint (Checkpoint): New Checkpoint to be updated
    
    Returns:
      New Checkpoint, hash that is used as fname
    '''
    curr_time = str(dt.datetime.now())
    fname_hash = self._generate_hash()
    new_checkpoint['datetime'] = curr_time
    return new_checkpoint, fname_hash
  
  def _save(self, updated_checkpoint, params_fname_hash):
    '''Saves the checkpoint

    Saves the checkpoint details onto checkpoints.json and creates a new file
    with the params of the model

    if self.session isn't already in existing checkpoints, then it creates a new dict
    and adds the checkpoint there.

    Args:
      updated_checkpoint (Checkpoint): Checkpoint that is updated
      params_fname_hash (str): Hash that is generated to be the name of filename
    '''
    with open(f'{self.dirpath}/checkpoints.json', 'r') as checkpoints_fp:
      existing_checkpoints = json.load(checkpoints_fp)
    with open(f'{self.dirpath}/checkpoints.json', 'w') as checkpoints_fp:
      if existing_checkpoints.get(self.session) is None:
        existing_checkpoints[self.session] = {}
      existing_checkpoints[self.session][params_fname_hash] = updated_checkpoint
      json.dump(existing_checkpoints, checkpoints_fp, indent=4)
    self.model.save(f'{self.dirpath}/{self.session}/{params_fname_hash}.hkl')
  
  def load(self, params_fname, load_params=True):
    '''Retrieves the Checkpoint

    Returns the checkpoint based on the params_fname and loads the params
    onto the model if load_params is True

    Args:
      params_fname (str): Filename to load params from
      load_params (bool): Whether params should be loaded from the file onto the model
    
    Returns:
      Checkpoint desired
    
    Raises:
      ValueError: If the current session is not present in checkpoints.py
    '''
    params_fname_hash = params_fname.rstrip('.hkl')
    with open(f'{self.dirpath}/checkpoints.json') as checkpoints_fp:
      session = json.load(checkpoints_fp).get(self.session)
      if session is None:
        raise ValueError(f"Invalid session {self.session}")
      if params_fname_hash not in session.keys():
        raise ValueError(f"File {params_fname} not in current session {self.session} directory! Please specify the session using Checkpoint.specify_session")
      checkpoint = session[params_fname_hash]
    if load_params:
      self.model.load(f'{self.dirpath}/{self.session}/{params_fname}')
    return checkpoint
  
  def _init_files(self, dirpath):
    '''Initializes files required for Checkpoint

    Creates a new folder at dirpath, if it doesn't exist
    A checkpoint.json file is created, if it is empty, then a new session is created
    if self.session is None, then automatically the last session is initialized as self.session

    Args:
      dirpath (str): Directory in which checkpoints must be saved
    '''
    dirpath = dirpath.rstrip("/'\'")
    try:
      os.mkdir(dirpath)
    except FileExistsError:
      pass
    self.dirpath = dirpath # All validation passed, can be assigned to self

    try:
      with open(f'{dirpath}/checkpoints.json', 'x') as checkpoints_fp: # Create checkpoints.json or do nothing if already exists
        pass
    except FileExistsError:
      pass
    
    with open(f'{dirpath}/checkpoints.json') as checkpoints_fp:
      contents = checkpoints_fp.read()
      if contents.strip()!='':
        sessions = json.loads(contents) #json.JSONDecodeError is raised if JSON file is invalid
        if self.session is None:
          if len(sessions)==0:
            self.new_session()
          else:
            self.session = list(sessions.keys())[-1]
      else: # checkpoints.json is empty
        sessions = {}
        self.new_session()
    with open(f'{dirpath}/checkpoints.json', 'w') as checkpoints_fp:
      json.dump(sessions, checkpoints_fp, indent=4)
  
  def _generate_hash(self):
    '''Generates 64 hex digit sha256 hash of a random number

    Returns:
      sha256 hash
    '''
    return sha256(secrets.token_hex(32).encode('utf-8')).hexdigest()[:self.hash_length]


from setuptools import setup, find_packages

def read_file(fpath):
  with open(fpath) as fp:
    data = fp.read()
  return data

setup(
  name = 'neograd',
  version = '0.0.3',
  author = 'Pranav Sastry',
  author_email = 'pranava.sri@gmail.com',
  maintainer = 'Pranav Sastry',
  url = 'https://github.com/pranftw/neograd',
  python_requires = '>=3.7',
  install_requires = read_file('requirements.txt').split('\n'),
  description = 'A deep learning framework created from scratch with Python and NumPy',
  license = 'GPL-3.0',
  keywords = 'python  ai  deep-learning  numpy automatic-differentiation  autograd neural-networks  pytorch-api',
  packages = find_packages(),
  long_description = read_file('README.md'),
  long_description_content_type='text/markdown'
)


import numpy as np
from ..autograd.utils import get_graph


class Optimizer:
  '''Base class for all optimizers

  Parameters:
    params (list of Param): Params that need to be updated
    lr (float): Learning rate
  '''
  def __init__(self, params, lr):
    self.params = params
    self.lr = lr

  def zero_grad(self, all_members=False):
    '''Resets the grads of tensors

    By default, since after loss.backward, only Tensors in memory are the params, only their
    gradients are reset since everytime a new graph is dynamically created

    However if retain_graph=True in backward, then all the members in the graph, need to be
    zero_grad-ed to get the correct gradients, to prevent this all_members can be set to True

    Args:
      all_members (bool): If all the members in the graph should be zero_grad-ed.
        Defaults to False
    '''
    if all_members:
      graph = get_graph()
      graph.zero_grad()
    for param in self.params: # This is done for redundancy, if all_members=True on a graph that's been reset
        param.zero_grad()


class GD(Optimizer):
  '''Vanilla Gradient Descent
  '''
  def __init__(self, params, lr):
    super().__init__(params, lr)
  
  def step(self):
    '''Updates the params
    '''
    for param in self.params:
      if param.requires_grad:
        param.data -= (self.lr*param.grad)
  
  def __repr__(self):
    return f'GD(params={self.params}, lr={self.lr})'
  
  def __str__(self):
    return f'GD(params={self.params}, lr={self.lr})'


class Momentum(Optimizer):
  '''Gradient Descent with Momentum
    
  https://youtu.be/k8fTYJPd3_I

  Parameters:
    beta (float): Value of Beta
  '''
  def __init__(self, params, lr, beta=0.9):
    super().__init__(params, lr)
    self.beta = beta
    self.init_momentum_grads()

  def step(self):
    '''Updates the params

    Uses the momentum_grads to update the params
    '''
    self.update_momentum_grads()
    for param in self.params:
      if param.requires_grad:
        param.data -= (self.lr*param.momentum_grad)
  
  def init_momentum_grads(self):
    '''Initializes momentum grads

    For each param, it sets its momentum_grad to 0
    '''
    for param in self.params:
      if param.requires_grad:
        param.momentum_grad = 0
  
  def update_momentum_grads(self):
    '''Updates momentum grads
    '''
    for param in self.params:
      if param.requires_grad:
        param.momentum_grad = (self.beta*param.momentum_grad) + ((1-self.beta)*param.grad)
  
  def __repr__(self):
    return f'Momentum(params={self.params}, lr={self.lr}, beta={self.beta})'
  
  def __str__(self):
    return f'Momentum(params={self.params}, lr={self.lr}, beta={self.beta})'


class RMSProp(Optimizer):
  '''RMSProp
    
  https://youtu.be/_e-LFe_igno

  Parameters:
    beta (float): Value of Beta
    epsilon (float): Value of epsilon
  '''
  def __init__(self, params, lr, beta=0.9, epsilon=1e-8):
    super().__init__(params, lr)
    self.beta = beta
    self.epsilon = epsilon
    self.init_rms_grads()

  def step(self):
    '''Updates the params

    Uses the rms_grads to update the params
    '''
    self.update_rms_grads()
    for param in self.params:
      if param.requires_grad:
        param.data -= (self.lr*(param.grad/(np.sqrt(param.rms_grad) + self.epsilon)))
  
  def init_rms_grads(self):
    '''Initializes rms grads

    For each param, it sets its rms_grad to 0
    '''
    for param in self.params:
      if param.requires_grad:
        param.rms_grad = 0
  
  def update_rms_grads(self):
    '''Updates rms grads
    '''
    for param in self.params:
      if param.requires_grad:
        param.rms_grad = (self.beta*param.rms_grad) + ((1-self.beta)*np.square(param.grad))
  
  def __repr__(self):
    return f'RMSProp(params={self.params}, lr={self.lr}, beta={self.beta}, epsilon={self.epsilon})'
  
  def __str__(self):
    return f'RMSProp(params={self.params}, lr={self.lr}, beta={self.beta}, epsilon={self.epsilon})'


class Adam(Optimizer):
  '''Adam

  https://youtu.be/JXQT_vxqwIs

  Parameters:
    iter (int): The number of iterations that has occurred, used for bias correction
    beta1 (float): Value of beta1
    beta2 (float): Value of beta2
    epsilon (float): Value of epsilon
  '''
  def __init__(self, params, lr, beta1=0.9, beta2=0.999, epsilon=1e-8):
    super().__init__(params, lr)
    self.iter = 0
    self.beta1, self.beta2 = beta1, beta2
    self.epsilon = epsilon
    self.init_adam_grads()
  
  def step(self):
    '''Updates the params

    Uses the rms_grads and momentum_grads to update the params
    '''
    self.iter+=1
    self.update_adam_grads()
    for param in self.params:
      if param.requires_grad:
        bias_corrected_momentum_grad = param.momentum_grad/(1-(self.beta1**self.iter))
        bias_corrected_rms_grad = param.rms_grad/(1-(self.beta2**self.iter))
        param.data -= (self.lr*(bias_corrected_momentum_grad/(np.sqrt(bias_corrected_rms_grad)+self.epsilon)))
  
  def init_adam_grads(self):
    '''Initializes rms grads and momentum grads

    For each param, it sets its rms_grad to 0, momentum_grad to 0
    '''
    for param in self.params:
      if param.requires_grad:
        param.momentum_grad = 0
        param.rms_grad = 0
  
  def update_adam_grads(self):
    '''Updates rms grads and momentum_grads
    '''
    for param in self.params:
      if param.requires_grad:
        param.momentum_grad = (self.beta1*param.momentum_grad) + ((1-self.beta1)*param.grad)
        param.rms_grad = (self.beta2*param.rms_grad) + ((1-self.beta2)*np.square(param.grad))

  def reset_iter(self):
    '''Resets iter to 0
    '''
    self.iter = 0
  
  def __repr__(self):
    return f'Adam(params={self.params}, lr={self.lr}, beta1={self.beta1}, beta2={self.beta2}, epsilon={self.epsilon})'
  
  def __str__(self):
    return f'Adam(params={self.params}, lr={self.lr}, beta1={self.beta1}, beta2={self.beta2}, epsilon={self.epsilon})'
    
  
